# Copyright cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http:# www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
from typing import Dict, List, Set, Callable
from dlk.core.base_module import SimpleModule, BaseModuleConfig
from . import decoder_register, decoder_config_register
from dlk.core.modules import module_config_register, module_register
import torch.nn as nn

@decoder_config_register("biaffine")
class BiAffineConfig(BaseModuleConfig):
    default_config = {
        "_name": "biaffine",
        "config": {
            "input_size": "*@*",
            "hidden_size": 0, # default equals to input_size
            "output_size": "*@*",
            "dropout": 0.0,
            "max_seq_len": 1024,
            "relation_position": False,
            "multi_matrix": 1,
            "output_map": {},
            "input_map": {}, # required_key: provide_key
        },
        "_link":{
            "config.input_size": ["module.config.input_size"],
            "config.hidden_size": ["module.config.hidden_size"],
            "config.output_size": ["module.config.output_size"],
            "config.max_seq_len": ["module.config.max_seq_len"],
            "config.relation_position": ["module.config.relation_position"],
            "config.multi_matrix": ["module.config.multi_matrix"],
            "config.dropout": ["module.config.dropout"],
        },
        "module": {
            "_base": "biaffine",
        },
    }
    """Config for BiAffine 
    """
    def __init__(self, config: Dict):
        super(BiAffineConfig, self).__init__(config)
        self.biaffine_config = config["module"]
        config = config['config']
        self.post_check(config, used=[
            "input_size",
            "output_size",
            "hidden_size",
            "multi_matrix",
            "dropout",
            "max_seq_len",
            "relation_position",
            "return_logits",
        ])


@decoder_register("biaffine")
class BiAffine(SimpleModule):
    """biaffine
    """
    def __init__(self, config: BiAffineConfig):
        super(BiAffine, self).__init__(config)
        self._provide_keys = {'logits'}
        self._required_keys = {'embedding'}
        self._provided_keys = set()

        self.config = config

        self.biaffine = module_register.get('biaffine')(module_config_register.get('biaffine')(config.biaffine_config))

    def init_weight(self, method: Callable):
        """init the weight of submodules by 'method'

        Args:
            method: init method

        Returns: 
            None

        """
        self.biaffine.init_weight(method)

    def forward(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        """

        Args:
            inputs: one mini-batch inputs

        Returns: 
            one mini-batch outputs

        """
        embedding = inputs[self.get_input_name('embedding')]
        inputs[self.get_output_name("logits")] = self.biaffine(embedding)
        if self._logits_gather.layer_map:
            inputs.update(self._logits_gather([inputs[self.get_output_name('logits')]]))
        return inputs



# Copyright 2021 cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hjson
import os
from typing import Dict, Union, Callable, List, Any
from dlk.utils.parser import BaseConfigParser
from dlk.utils.config import ConfigTool
from dlk.data.datamodules import datamodule_register, datamodule_config_register
from dlk.managers import manager_register, manager_config_register
from dlk.core.imodels import imodel_register, imodel_config_register
import pickle as pkl
from dlk.utils.io import open
import torch
import copy
import uuid
import json
from dlk.utils.logger import Logger

logger = Logger.get_logger()


class Predict(object):
    """Predict

    Config Example:
        >>> {
        >>>     "_focus": {
        >>>     },
        >>>     "_link": {},
        >>>     "_search": {},
        >>>     "config": {
        >>>         "save_dir": "*@*",  # must be provided
        >>>         "data_path": "*@*",  # must be provided
        >>>     },
        >>>     "task": {
        >>>         "_name": task_name
        >>>         ...
        >>>     }
        >>> }
    """
    def __init__(self, config: Union[str, dict], checkpoint: str):
        super(Predict, self).__init__()
        if not isinstance(config, dict):
            with open(config) as f:
                config = hjson.load(f, object_pairs_hook=dict)

        self.focus = config.pop('_focus', {})
        configs = BaseConfigParser(config).parser_with_check()
        assert len(
            configs
        ) == 1, f"For predict currently the config length must be 1(you cannot use _search in predict)."
        self.config = configs[0]

        # TODO: FIXME: use pytorch-lightning build in remote filesystem
        # https://pytorch-lightning.readthedocs.io/en/latest/common/remote_fs.html
        if isinstance(checkpoint, str):
            with open(checkpoint, 'rb') as f:
                self.ckpt = torch.load(f, map_location=torch.device('cpu'))
        else:
            self.ckpt = torch.load(checkpoint,
                                   map_location=torch.device('cpu'))
        config_name = []
        for source, to in self.focus.items():
            config_point = self.config
            trace = source.split('.')
            for t in trace:
                config_point = config_point[t]
            config_name.append(to + str(config_point))
        if config_name:
            name_str = '_'.join(config_name)
        else:
            name_str = self.config['root']['_name']
        self.name_str = name_str

    def convert2script(self, data=None):
        """trace the model to torchscript
        Returns: 
            TODO

        """
        config = self.config['root']
        name = self.name_str
        # get data
        if not data:
            data = self.get_data(config)

        # set datamodule
        datamodule = self.get_datamodule(config, data)

        # init imodel and inject the origin test and valid data
        imodel = self.get_imodel(config, data)

        dataloader = datamodule.train_dataloader()
        for data in dataloader:
            # script = torch.jit.trace(imodel.model, example_inputs=data, strict=False)
            # script = torch.jit.trace(imodel.model,
            #                          example_inputs=data,
            #                          strict=False)
            script = torch.jit.script(imodel.model, example_inputs=data)
            print(script)
            print(script(data))
            # imodel.model(data)
            break
        logger.error('The trace method is not implement yet.')
        raise NotImplementedError

    def predict(self, data=None, save_condition=False):
        """init the model, datamodule, manager then predict the predict_dataloader

        Args:
            data: if provide will not load from data_path

        Returns: 
            None

        """
        config = copy.deepcopy(self.config['root'])
        name = self.name_str
        # get data
        if not data:
            data = self.get_data(config)

        # set datamodule
        datamodule = self.get_datamodule(config, data)

        # set training manager
        manager = self.get_manager(config, name)

        # init imodel and inject the origin test and valid data
        imodel = self.get_imodel(config, data)

        # start predict
        predict_result = manager.predict(model=imodel, datamodule=datamodule)
        return imodel.postprocessor(stage='predict',
                                    list_batch_outputs=predict_result,
                                    origin_data=data['predict'],
                                    rt_config={},
                                    save_condition=save_condition)

    def get_data(self, config):
        """get the data decided by config

        Args:
            config: {"config": {"data_path": '..'}}

        Returns: 
            loaded data

        """

        with open(config['config']['data_path'], 'rb') as f:
            self.data = pkl.load(f).get('data', {})
        return self.data

    def get_datamodule(self, config, data):
        """get the datamodule decided by config, and fit the data to datamodule

        Args:
            config: {"task": {"datamodule": '..'}}
            data: {"train": '..', 'valid': '..', ..}

        Returns: 
            datamodule

        """
        DataModule, DataModuleConfig = ConfigTool.get_leaf_module(
            datamodule_register, datamodule_config_register, 'datamodule',
            config['task']['datamodule'])
        datamodule = DataModule(DataModuleConfig, data)
        return datamodule

    def get_manager(self, config, name):
        """get the tranin/predict manager decided by config

        Args:
            config: {"task": {"manager": '..'}, "config": {"save_dir"}}
            name: the predict progress name

        Returns: 
            manager

        """
        Manager, ManagerConfig = ConfigTool.get_leaf_module(
            manager_register, manager_config_register, 'manager',
            config.get('task').get('manager'))
        manager = Manager(ManagerConfig,
                          rt_config={
                              "save_dir": config.get('config').get("save_dir"),
                              "name": name
                          })
        return manager

    def get_imodel(self, config, data):
        """get the imodel decided by config

        Args:
            config: {"task": {"imodel": '..'}}
            data: {"train": '..', 'valid': '..', ..}

        Returns: 
            imodel

        """
        IModel, IModelConfig = ConfigTool.get_leaf_module(
            imodel_register, imodel_config_register, 'imodel',
            config.get('task').get('imodel'))
        imodel = IModel(IModelConfig, checkpoint=True)
        imodel.load_state_dict(self.ckpt['state_dict'])
        imodel.eval()
        return imodel



# Copyright cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dlk.utils.tokenizer_util import PreTokenizerFactory, TokenizerPostprocessorFactory, TokenizerNormalizerFactory
from dlk.utils.config import ConfigTool, BaseConfig
from typing import Dict, Callable, Union
import json
from dlk.utils.logger import Logger

from dlk.data.subprocessors import subprocessor_register, subprocessor_config_register, ISubProcessor
from tokenizers import normalizers
from tokenizers import pre_tokenizers

import pandas as pd
from tokenizers import Tokenizer
from tokenizers.models import WordPiece

logger = Logger.get_logger()


@subprocessor_config_register('fast_tokenizer')
class FastTokenizerConfig(BaseConfig):
    default_config = {
           "_name": "fast_tokenizer",
           "config": {
               "train": {
                   "data_set": {                   # for different stage, this processor will process different part of data
                       "train": ["train", "valid", 'test'],
                       "predict": ["predict"],
                       "online": ["online"]
                   },
                   "config_path": "*@*",
                   "truncation": {     # if this is set to None or empty, will not do trunc
                       "direction": "right", # default `right`, if set `left`, will reserve the rightest chars.
                       "stride": 0, # if the sequence is very long, will split to multiple span, stride is the window slide
                       "max_length": 512,
                       "strategy": "longest_first", # Can be one of longest_first, only_first or only_second.
                   },
                   "normalizer": "default", # ["nfd", "lowercase", "strip_accents", "some_processor_need_config": {config}], # if don't set this, will use the default normalizer from config
                   "pre_tokenizer": "default",# [{"whitespace": {}}], # if don't set this, will use the default normalizer from config
                   "post_processor": "default", # "bert", # if don't set this, will use the default normalizer from config, WARNING: not support disable  the default setting( so the default tokenizer.post_tokenizer should be null and only setting in this configure)
                   "output_map": { # this is the default value, you can provide other name
                       "tokens": "tokens",
                       "ids": "input_ids",
                       "attention_mask": "attention_mask",
                       "type_ids": "type_ids",
                       "special_tokens_mask": "special_tokens_mask",
                       "offsets": "offsets",
                       "word_ids": "word_ids",
                       "sequence_ids": "sequence_ids",
                   }, # the tokenizer output(the key) map to the value
                   "input_map": {
                       "sentence": "sentence", # for sigle input, tokenizer the "sentence"
                       "sentence_a": "sentence_a", # for pair inputs, tokenize the "sentence_a" && "sentence_b"
                       "sentence_b": "sentence_b", # for pair inputs
                       "pretokenized_words": "pretokenized_words", # pretokenized word related to sentence
                       "pretokenized_words_a": "pretokenized_words_a", # pretokenized word b related to sentence_a
                       "pretokenized_words_b": "pretokenized_words_b", # pretokenized word b related to sentence_b
                       "pretokenized_word_offsets": "pretokenized_word_offsets", # pretokenized word offsets for fix offset
                       "pretokenized_word_offsets_a": "pretokenized_word_offsets_a", # pretokenized word offsets for fix offset
                       "pretokenized_word_offsets_b": "pretokenized_word_offsets_b", # pretokenized word offsets for fix offset
                   },
                   "deliver": "tokenizer",
                   "process_data": { "is_pretokenized": False, "add_special_tokens": True},
                   "expand_examples": True, # if the sequence is very long, will split to multiple span, whether expand the examples
                   "data_type": "single", # single or pair, if not provide, will calc by len(process_data)
                   "fix_offset": False, # whether fix the offset for pretokenizerd word
               },
               "predict": ["train", {"deliver": None}],
               "extend_train": ["train", {"deliver": None}],
               "online": ["train", {"deliver": None}],
           }
       }
    """Config for FastTokenizer

    Config Example:
        default_config
    """
    def __init__(self, stage, config):
        super(FastTokenizerConfig, self).__init__(config)
        self.config = ConfigTool.get_config_by_stage(stage, config)
        self.data_set = self.config.get('data_set', {}).get(stage, [])
        if not self.data_set:
            return
        self.config_path = self.config.get('config_path')
        self.normalizer = self.config.get('normalizer', "default")
        self.pre_tokenizer = self.config.get('pre_tokenizer', "default")
        self.post_processor = self.config.get('post_processor', "default")
        self.expand_examples = self.config.get('expand_examples', False)
        self.truncation = self.config["truncation"]
        self.deliver = self.config['deliver']
        self.load = self.config.get('load', None)
        self.input_map = self.config['input_map']
        self.output_map = self.config.get('output_map', { # default
            "tokens": "tokens",
            "ids": "input_ids",
            "attention_mask": "attention_mask",
            "type_ids": "type_ids",
            "special_tokens_mask": "special_tokens_mask",
            "offsets": "offsets",
            "word_ids": "word_ids",
            "sequence_ids": "sequence_ids",
        })
        self.process_data = self.config['process_data']
        self.is_pretokenized = self.process_data['is_pretokenized']
        self.data_type = self.config["data_type"]
        self.fix_offset = self.config['fix_offset']
        assert self.data_type in ['single', 'pair']
        self.post_check(self.config, used=[
            "data_set",
            "config_path",
            "truncation",
            "normalizer",
            "pre_tokenizer",
            "post_processor",
            "output_map",
            "input_map",
            "fix_offset",
            "deliver",
            "process_data",
            "expand_examples",
            "data_type",
        ])


@subprocessor_register('fast_tokenizer')
class FastTokenizer(ISubProcessor):
    """ FastTokenizer use hugingface tokenizers

    Tokenizer the single $sentence
    Or tokenizer the pair $sentence_a, $sentence_b
    Generator $tokens, $input_ids, $type_ids, $special_tokens_mask, $offsets, $word_ids, $overflowing, $sequence_ids
    """

    def __init__(self, stage: str, config: FastTokenizerConfig):
        super().__init__()
        self.stage = stage
        self.config = config
        if not self.config.data_set:
            logger.info(f"Skip 'fast_tokenizer' at stage {self.stage}")
            return
        self.tokenizer = Tokenizer.from_file(self.config.config_path)
        pretokenizer_factory = PreTokenizerFactory(self.tokenizer)
        tokenizer_postprocessor_factory = TokenizerPostprocessorFactory(self.tokenizer)
        tokenizer_normalizer_factory = TokenizerNormalizerFactory(self.tokenizer)

        if not self.config.pre_tokenizer:
            self.tokenizer.pre_tokenizer = pre_tokenizers.Sequence([])
        elif self.config.pre_tokenizer != "default":
            assert isinstance(config.pre_tokenizer, list)
            pre_tokenizers_list = []
            for one_pre_tokenizer in config.pre_tokenizer:
                pre_tokenizers_list.append(self._get_processor(pretokenizer_factory, one_pre_tokenizer))
            self.tokenizer.pre_tokenizer = pre_tokenizers.Sequence(pre_tokenizers_list)

        if not self.config.post_processor:
            raise KeyError("The tokenizer is not support disable default tokenizers post processer. (You can delete the config manully)")
        elif self.config.post_processor != "default":
            self.tokenizer.post_processor = self._get_processor(tokenizer_postprocessor_factory, config.post_processor)

        if not self.config.normalizer:
            self.tokenizer.normalizer = normalizers.Sequence([])
        elif self.config.normalizer != "default":
            assert isinstance(config.normalizer, list)
            normalizers_list = []
            for one_normalizer in config.normalizer:
                normalizers_list.append(self._get_processor(tokenizer_normalizer_factory, one_normalizer))
            self.tokenizer.normalizer = normalizers.Sequence(normalizers_list)

        if self.config.truncation:
            self.tokenizer.enable_truncation(max_length=self.config.truncation['max_length'], 
                                             stride=self.config.truncation["stride"], 
                                             strategy=self.config.truncation['strategy'],
                                             direction=self.config.truncation["direction"])

    def _get_processor(self, factory: Union[PreTokenizerFactory, TokenizerNormalizerFactory, TokenizerPostprocessorFactory], one_processor: Union[Dict, str]):
        """return the processor in factory by the processor name and update the config of the processor if provide

        Args:
            factory: process factory
            one_processor: the processor info, it's name (and config)

        Returns: 
            processor

        """
        if isinstance(one_processor, dict):
            assert len(one_processor) == 1
            process_name, process_config = list(one_processor.items())[0]
            return factory.get(process_name)(**process_config)
        else:
            assert isinstance(one_processor, str)
            return factory.get(one_processor)()

    def _process(self, data: pd.DataFrame)->pd.DataFrame:
        """use self._tokenize tokenize the data

        Args:
            data: several data in dataframe

        Returns: 
            updated dataframe

        """
        if self.config.data_type=='single':
            batch_encodes = self.tokenizer.encode_batch(data[self.config.input_map['pretokenized_words']] if self.config.is_pretokenized else self.config.input_map['sentence'], **self.config.process_data)
        else: # pair
            sentence_as = data[self.config.input_map['pretokenized_words_a']] if self.config.is_pretokenized else data[self.config.input_map['sentence_a']]
            sentence_bs = data[self.config.input_map['pretokenized_words_b']] if self.config.is_pretokenized else data[self.config.input_map['sentence_b']]
            batch_encodes = self.tokenizer.encode_batch([(sentence_a, sentence_b) for sentence_a, sentence_b in zip(sentence_as, sentence_bs)], **self.config.process_data)
        if self.config.expand_examples:
            encodes_list = []
            for encode in batch_encodes:
                encodes_list.append([encode]+encode.overflowing)
            data['_tokenizer_encoders'] = encodes_list
            data = data.explode('_tokenizer_encoders', ignore_index=True)
        else:
            data['_tokenizer_encoders'] = batch_encodes
        tokens_list, ids_list, attention_mask_list, type_ids_list, special_tokens_mask_list, offsets_list, word_ids_list, overflowing_list, sequence_ids_list = [], [], [], [], [], [], [], [], []
        for encode in data['_tokenizer_encoders']:
            tokens_list.append(encode.tokens)
            ids_list.append(encode.ids)
            attention_mask_list.append(encode.attention_mask)
            type_ids_list.append(encode.type_ids)
            special_tokens_mask_list.append(encode.special_tokens_mask)
            offsets_list.append(encode.offsets)
            word_ids_list.append(encode.word_ids)
            # overflowing_list.append(encode.overflowing)
            sequence_ids_list.append(encode.sequence_ids)
        output_map = self.config.output_map
        data[output_map['tokens']] = tokens_list
        data[output_map['ids']] = ids_list
        data[output_map['attention_mask']] = attention_mask_list
        data[output_map['type_ids']] = type_ids_list
        data[output_map['special_tokens_mask']] = special_tokens_mask_list
        data[output_map['offsets']] = offsets_list
        data[output_map['word_ids']] = word_ids_list
        # data[output_map['overflowing']] = overflowing_list
        data[output_map['sequence_ids']] = sequence_ids_list
        data.drop("_tokenizer_encoders", axis=1, inplace=True)

        if self.config.is_pretokenized and self.config.fix_offset:
            data[output_map['offsets']] = data.apply(self._fix_offset, axis=1)
        return data

    def _fix_offset(self, one_line: pd.Series):
        """fix the pretokenizerd offset

        Args:
            one_line: a Series which contains the config.input_map['pretokenized_word_offsets'], config.output_map['offsets'], config.output_map['word_ids'], configs.output_map['type_ids']

        Returns: 
            encode.tokens, encode.ids, encode.attention_mask, encode.type_ids, encode.special_tokens_mask, encode.offsets, encode.word_ids, encode.overflowing, encode.sequence_ids 

        """
        fixed_offsets = []
        word_offsets = []
        if self.config.data_type == 'single':
            word_offsets = [one_line[self.config.input_map['pretokenized_word_offsets']]]
        else: # pair
            word_offsets = [one_line[self.config.input_map['pretokenized_word_offsets_a']], one_line[self.config.input_map['pretokenized_word_offsets_b']]]
        for offset, word_id, type_id in zip(one_line['offsets'], one_line['word_ids'], one_line['type_ids']):
            if offset == (0, 0):
                fixed_offsets.append(offset)
            else:
                fixed_offsets.append((offset[0]+word_offsets[type_id][word_id][0], offset[1]+word_offsets[type_id][word_id][0]))
        return fixed_offsets

    def process(self, data: Dict)->Dict:
        """Tokenizer entry

        Args:
            data: 
            >>> {
            >>>     "data": {"train": ...},
            >>>     "tokenizer": ..
            >>> }

        Returns: 
            data and the tokenizer info is in the data['data'], if you set the self.config.deliver, the data[self.config.deliver] will set to self.tokenizer.to_str()

        """
        if not self.config.data_set:
            return data
        for data_set_name in self.config.data_set:
            if data_set_name not in data['data']:
                logger.info(f'The {data_set_name} not in data. We will skip tokenize it.')
                continue
            data_set = data['data'][data_set_name]
            data_set = self._process(data_set)
            data['data'][data_set_name] = data_set
        if self.config.deliver:
            data[self.config.deliver] = self.tokenizer.to_str()
        return data



# Copyright 2021 cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dlk.utils.config import ConfigTool, BaseConfig
from dlk.utils.logger import Logger
from typing import Dict, Callable, Set, List
from dlk.data.subprocessors import subprocessor_register, subprocessor_config_register, ISubProcessor
import pickle as pkl
from dlk.utils.io import open
import copy
import os

logger = Logger.get_logger()


@subprocessor_config_register('save')
class SaveConfig(BaseConfig):
    default_config = {
        "_name": "save",
        "config":{
            "base_dir": "",
            "train":{
                "processed": "processed_data.pkl", # all data without meta
                "meta": "*@*"
            },
            "predict": {
                "processed": "processed_data.pkl",
            },
            "extend_train": {
                "processed": "processed_data_extend.pkl",
            },
        }
    }
    """Config for Save
    Config Example:
        default_config
    """
    def __init__(self, stage, config):
        super(SaveConfig, self).__init__(config)
        self.config = ConfigTool.get_config_by_stage(stage, config)
        self.base_dir: str = config.get('config').get("base_dir", "")
        self.post_check(self.config, used=[
            "processed",
            "meta",
        ])


@subprocessor_register('save')
class Save(ISubProcessor):
    """
    Save the processed data to $base_dir/$processed
    Save the meta data(like vocab, embedding, etc.) to $base_dir/$meta
    """
    def __init__(self, stage: str, config: SaveConfig):
        super().__init__()
        self.stage = stage
        self.config = config.config
        if not self.config:
            logger.info(f"Skip 'save' at stage {self.stage}")
            return
        self.base_dir = config.base_dir

    def save(self, data, path: str):
        """save data to path

        Args:
            data: pickleable data
            path: the path to data

        Returns: 
            loaded data

        """
        logger.info(f"Saving file to {os.path.join(self.base_dir, path)}")
        with open(os.path.join(self.base_dir, path), 'wb') as f:
            return pkl.dump(data, f)

    def process(self, data: Dict) -> Dict:
        """Save entry

        Args:
            data: 
            >>> {
            >>>     "data": {"train": ...},
            >>>     "tokenizer": ..
            >>> }

        Returns: 
            data

        """
        if not self.config:
            return data
        # record for save only in meta, not in processed_data but returned all
        meta_fileds = set()
        if "meta" in self.config:
            for save_path, save_fileds in self.config['meta'].items():
                assert isinstance(save_fileds, list)
                meta_data = {}
                for field in save_fileds:
                    meta_data[field] = copy.deepcopy(data[field])
                    meta_fileds.add(field)
                self.save(meta_data, save_path)
        reserve_meta_map = {}
        for filed in meta_fileds:
            reserve_meta_map[filed] = data.pop(filed)
        if "processed" in self.config:
            self.save(data, self.config['processed'])
        data.update(reserve_meta_map)
        return data



# Copyright cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Callable, Dict, Any

__version__ = '0.0.15'


class AdditionalRegister(object):
    """Register"""
    def __init__(self, register_name:str):
        super(AdditionalRegister, self).__init__()
        self.register_name = register_name
        self.registry:Dict[str, Any] = {
            "multi_loss": {},
        }

    def register(self, group, name: str='')->Callable:
        """register the name: module to self.registry

        Args:
            group: the registed group name
            name: the registed function name

        Returns: 
            the module

        """
        if group not in self.registry:
            raise ValueError(f'Cannot recognize group {group}')
        register_group = self.registry[group]
        def decorator(module):
            if name.strip() == "":
                raise ValueError(f'You must set a name for {module.__name__}')

            if name in register_group:
                raise ValueError(f'The {group}/{name} is already registed in {self.register_name}.')
            register_group[name] = module
            self.registry[group] = register_group
            return module
        return decorator

    def __call__(self, group, name:str="")->Callable:
        """you can directly call the object, the behavior is the same as object.register(name)
        """
        return self.register(group, name)

    def get(self, group, name: str='')->Any:
        """get the module by name

        Args:
            group: the registed group name
            name: the name should be the real name or name+@+sub_name, and the

        Returns: 
            registed module

        """
        if group not in self.registry:
            raise KeyError(f"In '{self.register_name}' register, there is not a entry named '{group}/{name}'")
        return self.registry[group][name]

    def __getitem__(self, group, name: str='')->Any:
        """wrap for object.get(name)
        """
        return self.get(group, name)


additional_function = AdditionalRegister("additional_register")



# Copyright cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List
import torch.nn as nn
from . import loss_register, loss_config_register
from dlk import additional_function
import torch.nn as nn
import torch
from dlk.utils.config import ConfigTool

@additional_function("multi_loss", "sum")
def loss_sum(losses: Dict[str, torch.Tensor], **args:Dict):
    """sum all losses

    Args:
        losses (List): list of loss

    Returns: 
        sum of losses
    """
    loss = sum([losses[key] for key in losses])
    return loss

@loss_config_register("multi_loss")
class MultiLossConfig(object):
    default_config = {
        "config": {
            "loss_collect": "sum",
            "args": {},
            "log_map": {
                "loss": "loss"
            },
        },
        "_name": "multi_loss",
    }
    """Config for MultiLoss

    Config Example:
    """
    def __init__(self, config: Dict):
        super(MultiLossConfig, self).__init__()
        self.loss_configs = {}
        self.module_rank = []
        self.loss_collect = additional_function.get("multi_loss", config['config']['loss_collect'])
        self.loss_collect_name = config['config']['loss_collect']
        self.log_map = config['config']['log_map']
        if isinstance(self.log_map, str):
            self.log_map = {"loss": self.log_map}
        for loss in config:
            if loss in {'config', "_name", "_base"}:
                continue
            self.module_rank.append(loss)
            module_class, module_config = ConfigTool.get_leaf_module(loss_register, loss_config_register, "loss", config[loss])
            self.loss_configs[loss] = {
                "loss_class": module_class,
                "loss_config": module_config,
            }

@loss_register("multi_loss")
class MultiLoss(nn.Module):
    """ This module is NotImplemented yet don't use it
    """
    def __init__(self, config: MultiLossConfig):
        super(MultiLoss, self).__init__()
        self.config = config
        self.loss_collect = config.loss_collect
        self.module_rank = config.module_rank
        self.losses = nn.ModuleDict({
            loss_name: config.loss_configs[loss_name]['loss_class'](config.loss_configs[loss_name]['loss_config'])
            for loss_name in config.module_rank
        })

    def update_config(self, rt_config):
        """callback for imodel to update the total steps and epochs

        when init the loss module, the total step and epoch is not known, when all data ready, the imodel update the value for loss module

        Args:
            rt_config: { "total_steps": self.num_training_steps, "total_epochs": self.num_training_epochs}

        Returns: 
            None

        """
        for _, loss_module in self.losses.items():
            loss_module.update_config(rt_config)

    def calc(self, result, inputs, rt_config):
        """calc the loss the predict is from result, the ground truth is from inputs

        Args:
            result: the model predict dict
            inputs: the all inputs for model
            rt_config: provide the current training status 
                >>> {
                >>>     "current_step": self.global_step,
                >>>     "current_epoch": self.current_epoch,
                >>>     "total_steps": self.num_training_steps,
                >>>     "total_epochs": self.num_training_epochs
                >>> }

        Returns: 
            loss

        """
        losses = {}
        log_loss = {}
        for loss_module in self.module_rank:
            loss, log = self.losses[loss_module](result, inputs, rt_config)
            losses[loss_module] = loss
            log_loss.update(log)
        loss = self.loss_collect(losses=losses, rt_config=rt_config)
        if self.config.loss_collect_name != 'sum':
            log_loss.update({self.config.log_map.get("sum_loss", "sum_loss"): sum([losses[key] for key in losses])})
        log_loss.update({self.config.log_map['loss']: loss})
        return loss, log_loss

    def __call__(self, result, inputs, rt_config):
        """same as self.calc
        """
        return self.calc(result, inputs, rt_config)



# Copyright 2021 cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dlk.utils.vocab import Vocabulary
from dlk.utils.config import BaseConfig, ConfigTool
from typing import Dict, Callable, Set, List
from dlk.data.subprocessors import subprocessor_register, subprocessor_config_register, ISubProcessor
from functools import partial
from dlk.utils.logger import Logger
import os

logger = Logger.get_logger()

@subprocessor_config_register('token2charid')
class Token2CharIDConfig(BaseConfig):
    default_config = {
        "_name": "token2charid",
        "config": {
            "train":{
                "data_pair": {
                    "sentence & offsets": "char_ids"
                },
                "data_set": {                   # for different stage, this processor will process different part of data
                    "train": ['train', 'valid', 'test', 'predict'],
                    "predict": ['predict'],
                    "online": ['online']
                },
                "vocab": "char_vocab", # usually provided by the "token_gather" module
                "max_token_len": 20, # the max length of token, then the output will be max_token_len x token_num (put max_token_len in previor is for padding on token_num)
            },
            "predict": "train",
            "extend_train": "train",
            "online": "train",
        }
    }
    """Config for Token2CharID

    Config Example:
        default_config
    """

    def __init__(self, stage, config: Dict):

        super(Token2CharIDConfig, self).__init__(config)
        self.config = ConfigTool.get_config_by_stage(stage, config)
        self.data_set = self.config.get('data_set', {}).get(stage, [])
        if not self.data_set:
            return
        self.data_pair = self.config.pop('data_pair', {})

        if self.data_set and (not self.data_pair):
            raise ValueError("You must provide 'data_pair' for token2charid.")
        self.vocab = self.config.get('vocab', "")
        if self.data_set and (not self.vocab):
            raise ValueError("You must provide 'vocab' for token2charid.")
        self.max_token_len = self.config['max_token_len']
        self.post_check(self.config, used=[
            "data_pair",
            "data_set",
            "vocab",
            "max_token_len",
        ])


@subprocessor_register('token2charid')
class Token2CharID(ISubProcessor):
    """Use 'Vocabulary' map the character from tokens to id
    """

    def __init__(self, stage: str, config: Token2CharIDConfig):
        super().__init__()
        self.stage = stage
        self.config = config
        self.data_set = config.data_set
        if not self.data_set:
            logger.info(f"Skip 'token2charid' at stage {self.stage}")
            return
        self.data_pair = config.data_pair


    def process(self, data: Dict)->Dict:
        """Token2CharID Entry

        one_token like 'apple' will generate [1, 2, 2, 3] if max_token_len==4 and the vocab.word2idx = {'a': 1, "p": 2, "l": 3}

        Args:
            data: will process data

        Returns: 
            updated data(token -> char_ids)

        """


        if not self.data_set:
            return data

        def get_index_wrap(sentence_name, offset_name, x):
            """wrap the vocab.get_index"""
            sentence = list(x[sentence_name])
            offsets = x[offset_name]
            char_ids = []
            for offset in offsets:
                token = sentence[offset[0]: offset[1]][:self.config.max_token_len]
                token = token + [vocab.pad] * (self.config.max_token_len-len(token))
                char_ids.append([vocab.get_index(c) for c in token])
            return char_ids

        vocab = Vocabulary.load(data[self.config.vocab])

        for data_set_name in self.data_set:
            if data_set_name not in data['data']:
                logger.info(f'The {data_set_name} not in data. We will skip do token2charid on it.')
                continue
            data_set = data['data'][data_set_name]
            for key, value in self.data_pair.items():
                sentence_name, offset_name = key.split('&')
                sentence_name = sentence_name.strip()
                offset_name = offset_name.strip()
                get_index = partial(get_index_wrap, sentence_name, offset_name)
                if os.environ.get('DISABLE_PANDAS_PARALLEL', 'false') != 'false':
                    data_set[value] = data_set.parallel_apply(get_index, axis=1)
                else:
                    data_set[value] = data_set.apply(get_index, axis=1)
        return data



# Copyright 2021 cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch.nn as nn
from . import callback_register, callback_config_register
from typing import Dict, List
import os
from pytorch_lightning.callbacks import StochasticWeightAveraging


@callback_config_register('weight_average')
class StochasticWeightAveragingCallbackConfig(object):
    """Config for StochasticWeightAveragingCallback

    Config Example:
        >>> {   //weight_average default
        >>>     "_name": "weight_average",
        >>>     "config": {
        >>>         "swa_epoch_start": 0.8, // swa start epoch
        >>>         "swa_lrs": null,
        >>>             //None. Use the current learning rate of the optimizer at the time the SWA procedure starts.
        >>>             //float. Use this value for all parameter groups of the optimizer.
        >>>             //List[float]. A list values for each parameter group of the optimizer.
        >>>         "annealing_epochs": 10,
        >>>         "annealing_strategy": 'cos',
        >>>         "device": null, // save device, null for auto detach, if the gpu is oom, you should change this to 'cpu'
        >>>     }
        >>> }
    """
    def __init__(self, config):
        super(StochasticWeightAveragingCallbackConfig, self).__init__()
        config = config['config']
        self.swa_epoch_start = config['swa_epoch_start']
        self.swa_lrs = config["swa_lrs"]
        self.annealing_epochs = config["annealing_epochs"]
        self.annealing_strategy = config["annealing_strategy"]
        self.device = config["device"]

@callback_register('weight_average')
class StochasticWeightAveragingCallback(object):
    """Average weight by config
    """

    def __init__(self, config: StochasticWeightAveragingCallbackConfig):
        super().__init__()
        self.config = config

    def __call__(self, rt_config: Dict)->StochasticWeightAveraging:
        """return StochasticWeightAveraging object

        Args:
            rt_config: runtime config, include save_dir, and the checkpoint path name

        Returns: 
            StochasticWeightAveraging object

        """
        return StochasticWeightAveraging(**self.config.__dict__)



# Copyright 2021 cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from transformers.models.bert.modeling_bert import BertModel
from transformers.models.bert.configuration_bert import BertConfig
import json
import os
import torch.nn as nn
import torch
from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence
from typing import Dict
from . import module_register, module_config_register, Module
from dlk.utils.config import BaseConfig
from dlk.utils.io import open


@module_config_register("bert")
class BertWrapConfig(BaseConfig):
    """Config for BertWrap

    Config Example:
        >>> {
        >>>     "config": {
        >>>         "pretrained_model_path": "*@*",
        >>>         "from_pretrain": true,
        >>>         "freeze": false,
        >>>         "dropout": 0.0,
        >>>     },
        >>>     "_name": "bert",
        >>> }
    """

    def __init__(self, config: Dict):
        super(BertWrapConfig, self).__init__(config)
        self.pretrained_model_path = config['config']['pretrained_model_path']
        self.from_pretrain = config['config']['from_pretrain']
        self.freeze = config['config']['freeze']
        self.dropout = config['config']['dropout']
        if os.path.isdir(self.pretrained_model_path):
            if os.path.exists(os.path.join(self.pretrained_model_path, 'config.json')):
                with open(os.path.join(self.pretrained_model_path, 'config.json'), 'r') as f:
                    self.bert_config = BertConfig(**json.load(f))
            else:
                raise PermissionError(f"config.json must in the dir {self.pretrained_model_path}")
        else:
            if os.path.isfile(self.pretrained_model_path):
                try:
                    with open(self.pretrained_model_path, 'r') as f:
                        self.bert_config = BertConfig(**json.load(f))
                except:
                    raise PermissionError(f"You must provide the pretrained model dir or the config file path.")
        self.post_check(config['config'], used=['pretrained_model_path', 'from_pretrain', 'freeze', 'dropout'])


@module_register("bert")
class BertWrap(Module):
    """Bert wrap"""
    def __init__(self, config: BertWrapConfig):
        super(BertWrap, self).__init__()
        self.config = config

        self.bert = BertModel(config.bert_config, add_pooling_layer=True)
        self.dropout = nn.Dropout(float(self.config.dropout))

    def init_weight(self, method):
        """init the weight of model by 'bert.init_weight()' or from_pretrain

        Args:
            method: init method, no use for pretrained_transformers

        Returns: 
            None

        """
        if self.config.from_pretrain:
            self.from_pretrained()
        else:
            self.bert.init_weights()

    def from_pretrained(self):
        """init the model from pretrained_model_path
        """
        self.bert = BertModel.from_pretrained(self.config.pretrained_model_path)

    def forward(self, inputs: Dict):
        """do forward on a mini batch

        Args:
            batch: a mini batch inputs

        Returns: 
            sequence_output, all_hidden_states, all_self_attentions

        """
        if self.config.freeze:
            with torch.no_grad():
                outputs = self.bert(
                    input_ids = inputs.get("input_ids", None),
                    attention_mask = inputs.get("attention_mask", None),
                    token_type_ids = inputs.get("token_type_ids", None),
                    position_ids = inputs.get("position_ids", None),
                    head_mask = inputs.get("head_mask", None),
                    inputs_embeds = inputs.get("inputs_embeds", None),
                    encoder_hidden_states = inputs.get("encoder_hidden_states", None),
                    encoder_attention_mask = inputs.get("encoder_attention_mask", None),
                    past_key_values = inputs.get("past_key_values", None),
                    use_cache = None,
                    output_attentions = True,
                    output_hidden_states = True,
                    return_dict = False
                )
        else:
            outputs = self.bert(
                input_ids = inputs.get("input_ids", None),
                attention_mask = inputs.get("attention_mask", None),
                token_type_ids = inputs.get("token_type_ids", None),
                position_ids = inputs.get("position_ids", None),
                head_mask = inputs.get("head_mask", None),
                inputs_embeds = inputs.get("inputs_embeds", None),
                encoder_hidden_states = inputs.get("encoder_hidden_states", None),
                encoder_attention_mask = inputs.get("encoder_attention_mask", None),
                past_key_values = inputs.get("past_key_values", None),
                use_cache = None,
                output_attentions = True,
                output_hidden_states = True,
                return_dict = False
            )
        assert len(outputs) == 4, f"Please check transformers version, the len(outputs) is 4 in version == 4.12|4.15, or check your config and remove the 'add_cross_attention'"
        sequence_output, all_hidden_states, all_self_attentions = outputs[0], outputs[2], outputs[3]
        sequence_output = self.dropout(sequence_output)
        return sequence_output, all_hidden_states, all_self_attentions



# Copyright 2021 cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from dlk.core.imodels import imodel_config_register, imodel_register
from dlk.core.models import model_config_register, model_register
from dlk.core.callbacks import callback_config_register, callback_register
from dlk.core.initmethods import initmethod_config_register, initmethod_register
from dlk.core.layers import embedding_register, embedding_config_register, decoder_config_register, decoder_register, encoder_config_register, encoder_register
from dlk.core.optimizers import optimizer_config_register, optimizer_register
from dlk.core.schedulers import scheduler_config_register, scheduler_register
from dlk.core.losses import loss_config_register, loss_register
from dlk.core.modules import module_config_register, module_register



# Copyright 2021 cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import smart_open
from dlk.utils.logger import Logger
from io import BytesIO
import os
logger = Logger.get_logger()


def open(path: str, *args, **kwargs):
    if len(args)>0 and 'w' in args[0] or 'w' in kwargs.get('mode', ''):
        subdir = os.path.dirname(path)
        if path.startswith("hdfs://"):
            pass
        else:
            if ':/' in subdir:
                logger.error(f"Currently dlk is only support `hdfs` and `local` file.")
                raise PermissionError
            if not os.path.exists(subdir):
                try:
                    os.makedirs(subdir, exist_ok=True)
                except Exception as e:
                    logger.error(f"Currently dlk is only support `hdfs` and `local` file.")
                    raise e
    # some binary data(hdfs) is not seekable
    if len(args)>0 and 'rb' in args[0] or 'rb' in kwargs.get('mode', ''):
        if len(args)>0:
            mode = args[0]
        else:
            mode = kwargs['mode']
        assert mode == 'rb', f"Currently dlk is only support 'rb' for reading binary data"
        with smart_open.open(path, mode) as f:
            b_data = f.read()
        return BytesIO(b_data)
        
    return smart_open.open(path, *args, **kwargs)



# Copyright 2021 cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""basic modules"""
import importlib
import os
from dlk.utils.register import Register
import torch.nn as nn
from typing import Dict
import torch

module_config_register = Register("Module config register.")
module_register = Register("Module register.")


class Module(nn.Module):
    """This class is means DLK Module for replace the torch.nn.Module in this project
    """

    def init_weight(self, method):
        """init the weight of submodules by 'method'

        Args:
            method: init method

        Returns: 
            None

        """
        for module in self.children():
            module.apply(method)

    def forward(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        """in simple module, all step fit to this method

        Args:
            inputs: one mini-batch inputs

        Returns: 
            one mini-batch outputs

        """
        raise NotImplementedError

    def predict_step(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        """do predict for one batch

        Args:
            inputs: one mini-batch inputs

        Returns: 
            one mini-batch outputs

        """
        return self(inputs)

    def training_step(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        """do train for one batch

        Args:
            inputs: one mini-batch inputs

        Returns: 
            one mini-batch outputs

        """
        return self(inputs)

    def validation_step(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        """do validation for one batch

        Args:
            inputs: one mini-batch inputs

        Returns: 
            one mini-batch outputs

        """
        return self(inputs)

    def test_step(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        """do test for one batch

        Args:
            inputs: one mini-batch inputs

        Returns: 
            one mini-batch outputs

        """
        return self(inputs)



def import_modules(modules_dir, namespace):
    for file in os.listdir(modules_dir):
        path = os.path.join(modules_dir, file)
        if (
            not file.startswith("_")
            and not file.startswith(".")
            and (file.endswith(".py") or os.path.isdir(path))
        ):
            module_name = file[: file.find(".py")] if file.endswith(".py") else file
            importlib.import_module(namespace + "." + module_name)


# automatically import any Python files in the modules directory
modules_dir = os.path.dirname(__file__)
import_modules(modules_dir, "dlk.core.modules")



import torch
import random
seed = random.randint(1e9, 1e10)
torch.manual_seed(seed) # NOTE: should fix manual seed for every forward

class FGM():
    """
    Fast Gradient Method, ICLR-2017
    Examples
    --------
    >>> fgm.attack(epsilon, 'word_embeddings')
    >>> loss_adv = model(**batch).loss
    >>> loss_adv = loss_adv / args.gradient_accumulation_steps
    >>> accelerator.backward(loss_adv)
    >>> fgm.restore('word_embeddings')
    >>> optimizer.step()
    References
    ----------
    1. https://zhuanlan.zhihu.com/p/91269728
    """
    def __init__(self, model):
        self.model = model
        self.backup = {}

    def attack(self, epsilon=1., emb_name='emb.'):
        for name, param in self.model.named_parameters():
            if param.requires_grad and emb_name in name:
                self.backup[name] = param.data.clone()
                norm = torch.norm(param.grad)
                if norm != 0:
                    r_at = epsilon * param.grad / norm
                    param.data.add_(r_at)

    def restore(self, emb_name='emb.'):
        for name, param in self.model.named_parameters():
            if param.requires_grad and emb_name in name: 
                assert name in self.backup
                param.data = self.backup[name]
        self.backup = {}


class PGD():
    """
    Projected Gradient Descent, ICLR-2018 
    Examples
    --------
    >>> pgd.backup_grad()
    >>> for t in range(K):
    >>>     pgd.attack(emb_name='word_embeddings', is_first_attack=(t==0))
    >>>     if t != K-1:
    >>>         optimizer.zero_grad()
    >>>     else:
    >>>         pgd.restore_grad()
    >>>     loss_adv = model(**batch).loss
    >>>     loss_adv = loss_adv / args.gradient_accumulation_steps
    >>>     accelerator.backward(loss_adv)
    >>> pgd.restore(emb_name='word_embeddings')
    References
    ----------
    1. https://zhuanlan.zhihu.com/p/91269728
    """
    def __init__(self, model):
        self.model = model
        self.emb_backup = {}
        self.grad_backup = {}

    def attack(self, epsilon=1., alpha=0.3, emb_name='emb.', is_first_attack=False):
        for name, param in self.model.named_parameters():
            if param.requires_grad and emb_name in name:
                if is_first_attack:
                    self.emb_backup[name] = param.data.clone()
                norm = torch.norm(param.grad)
                if norm != 0:
                    r_at = alpha * param.grad / norm
                    param.data.add_(r_at)
                    param.data = self.project(name, param.data, epsilon)

    def restore(self, emb_name='emb.'):
        for name, param in self.model.named_parameters():
            if param.requires_grad and emb_name in name: 
                assert name in self.emb_backup
                param.data = self.emb_backup[name]
        self.emb_backup = {}
        
    def project(self, param_name, param_data, epsilon):
        r = param_data - self.emb_backup[param_name]
        if torch.norm(r) > epsilon:
            r = epsilon * r / torch.norm(r)
        return self.emb_backup[param_name] + r
        
    def backup_grad(self):
        for name, param in self.model.named_parameters():
            if param.requires_grad:
                self.grad_backup[name] = param.grad.clone()
    
    def restore_grad(self):
        for name, param in self.model.named_parameters():
            if param.requires_grad:
                param.grad = self.grad_backup[name]


class FreeLB(object):
    """
    Free Large-Batch, ICLR-2020
    Examples
    --------
    >>> # FreeLB training
    >>> freelb.backup_grad()
    >>> for t in range(K):
    >>>     freelb.attack(emb_name='word_embeddings', is_first_attack=(t==0))
    >>>     if t == 0:
    >>>         optimizer.zero_grad()
    >>>     loss_adv = model(**batch).loss
    >>>     loss_adv = loss_adv / K
    >>>     loss_adv = loss_adv / args.gradient_accumulation_steps  
    >>>     accelerator.backward(loss_adv)
    >>> freelb.restore_grad()
    >>> freelb.restore(emb_name='word_embeddings')
    >>> optimizer.step()
    >>> # Smart training (Smart Without Smoothness-Inducing Adversarial Regularization)
    >>> freelb.backup_grad()
    >>> for t in range(K):                    
    >>>     freelb.attack(emb_name='word_embeddings', is_first_attack=(t==0))
    >>>     if t == 0:
    >>>         optimizer.zero_grad()
    >>>     loss_adv,logits_adv = model(**batch)[:2]
    >>>     loss_adv = loss_adv / K 
    >>>     loss_adv = loss_adv / args.gradient_accumulation_steps
    >>>     accelerator.backward(loss_adv)
    >>> freelb.restore_grad()
    >>> ###Bregman Proximal Point Optimization (prevent the model from aggressive update)
    >>> loss_adv_last,logits_adv_last = model(**batch)[:2]
    >>> loss_kl = stable_kl(logits_adv_last, logits.detach(), reduce=False)
    >>> loss_kl = loss_kl / args.gradient_accumulation_steps
    >>> accelerator.backward(loss_kl * breg_miu)
    >>> freelb.restore(emb_name='word_embeddings')
    >>> optimizer.step()
    References
    ----------
    1. https://github.com/cuixuage/tf_bert_competitions/tree/7c39ae157de2c716c7c1eafbe67b143107008650
    """
    def __init__(self, model):
        self.model = model
        self.emb_backup = {}
        self.grad_backup = {}

    def attack(self, epsilon=1., alpha=0.3, emb_name='emb.', is_first_attack=False):
        for name, param in self.model.named_parameters():
            if param.requires_grad and emb_name in name:
                norm = torch.norm(param.grad)
                if is_first_attack:
                    self.emb_backup[name] = param.data.clone()
                if norm != 0:
                    r_at = alpha * param.grad / norm
                    param.data.add_(r_at)
                    param.data = self.project(name, param.data, epsilon)

    def restore(self, emb_name='emb.'):
        for name, param in self.model.named_parameters():
            if param.requires_grad and emb_name in name: 
                assert name in self.emb_backup
                param.data = self.emb_backup[name]
        self.emb_backup = {}

    def project(self, param_name, param_data, epsilon):
        r = param_data - self.emb_backup[param_name]
        if torch.norm(r) > epsilon:
            r = epsilon * r / torch.norm(r)
        return self.emb_backup[param_name] + r

    def backup_grad(self):
        for name, param in self.model.named_parameters():
            if param.requires_grad:
                self.grad_backup[name] = param.grad.clone()
    
    def restore_grad(self):
        for name, param in self.model.named_parameters():
            if param.requires_grad:
                param.grad = param.grad + self.grad_backup[name]



# Copyright 2021 cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch.nn as nn
from . import callback_register, callback_config_register
from typing import Dict, List
import os
from pytorch_lightning.callbacks import EarlyStopping


@callback_config_register('early_stop')
class EarlyStoppingCallbackConfig(object):
    """Config for EarlyStoppingCallback

    Config Example:
        >>> {
        >>>     "_name": "early_stop",
        >>>     "config":{
        >>>         "monitor": "val_loss",
        >>>         "mode": "*@*", // min or max, min for the monitor is loss, max for the monitor is acc, f1, etc.
        >>>         "patience": 3,
        >>>         "min_delta": 0.0,
        >>>         "check_on_train_epoch_end": null,
        >>>         "strict": true, // if the monitor is not right, raise error
        >>>         "stopping_threshold": null, // float, if the value is good enough, stop
        >>>         "divergence_threshold": null, // float,  if the value is so bad, stop
        >>>         "verbose": true, //verbose mode print more info
        >>>     }
        >>> }
    """
    def __init__(self, config: Dict):
        super(EarlyStoppingCallbackConfig, self).__init__()
        config = config['config']
        self.monitor = config['monitor']
        self.mode = config['mode']
        self.patience = config["patience"]
        self.min_delta = config['min_delta']
        self.strict = config['strict']
        self.verbose = config['verbose']
        self.stopping_threshold = config['stopping_threshold']
        self.divergence_threshold = config['divergence_threshold']
        self.check_on_train_epoch_end = config["check_on_train_epoch_end"]

@callback_register('early_stop')
class EarlyStoppingCallback(object):
    """Early stop decided by config
    """

    def __init__(self, config: EarlyStoppingCallbackConfig):
        super().__init__()
        self.config = config

    def __call__(self, rt_config: Dict)->EarlyStopping:
        """return EarlyStopping object

        Args:
            rt_config: runtime config, include save_dir, and the checkpoint path name

        Returns: 
            EarlyStopping object

        """
        return EarlyStopping(**self.config.__dict__)



# Copyright cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pickle as pkl
import json
from typing import Dict, List, Optional, Tuple, Union
import os
import numpy as np
import pandas as pd
import torch
from typing import Dict
from dlk.data.postprocessors import postprocessor_register, postprocessor_config_register, IPostProcessor, IPostProcessorConfig
from dlk.utils.logger import Logger
from dlk.utils.vocab import Vocabulary
from tokenizers import Tokenizer
from dlk.utils.io import open
import torchmetrics
logger = Logger.get_logger()


@postprocessor_config_register('multi_post')
class MultiPostPostProcessorConfig(IPostProcessorConfig):
    default_config = {
        "_name": "multi_post",
        "config": {
        }
    }
    """Config for MultiPostPostProcessor

    Config Example:
    """

    def __init__(self, config: Dict):
        super(MultiPostPostProcessorConfig, self).__init__(config)

        self.save_path = self.config['save_path']
        self.save_root_path = self.config['save_root_path']
        self.start_save_epoch = self.config['start_save_epoch']
        self.start_save_step = self.config['start_save_step']

        self.post_check(self.config, used=[
            "save_root_path",
            "save_path",
            "start_save_step",
            "start_save_epoch",
        ])


class AggregationStrategy(object):
    """docstring for AggregationStrategy"""
    NONE = "none"
    SIMPLE = "simple"
    FIRST = "first"
    AVERAGE = "average"
    MAX = "max"


@postprocessor_register('multi_post')
class MultiPostPostProcessor(IPostProcessor):
    """PostProcess for sequence labeling task"""
    def __init__(self, config: MultiPostPostProcessorConfig):
        super(MultiPostPostProcessor, self).__init__(config)

    def do_predict(self, stage: str, list_batch_outputs: List[Dict], origin_data: pd.DataFrame, rt_config: Dict)->List:
        """Process the model predict to human readable format

        There are three predictor for diffrent multi_post task dependent on the config.use_crf(the predict is already decoded to ids), and config.word_ready(subword has gathered to firstpiece)

        Args:
            stage: train/test/etc.
            list_batch_outputs: a list of outputs
            origin_data: the origin pd.DataFrame data, there are some data not be able to convert to tensor
            rt_config:
                >>> current status
                >>> {
                >>>     "current_step": self.global_step,
                >>>     "current_epoch": self.current_epoch,
                >>>     "total_steps": self.num_training_steps,
                >>>     "total_epochs": self.num_training_epochs
                >>> }

        Returns: 
            all predicts

        """
        raise NotImplementedError
        predicts = []
        return predicts

    def do_calc_metrics(self, predicts: List, stage: str, list_batch_outputs: List[Dict], origin_data: pd.DataFrame, rt_config: Dict)->Dict:
        """calc the scores use the predicts or list_batch_outputs

        Args:
            predicts: list of predicts
            stage: train/test/etc.
            list_batch_outputs: a list of outputs
            origin_data: the origin pd.DataFrame data, there are some data not be able to convert to tensor
            rt_config:
                >>> current status
                >>> {
                >>>     "current_step": self.global_step,
                >>>     "current_epoch": self.current_epoch,
                >>>     "total_steps": self.num_training_steps,
                >>>     "total_epochs": self.num_training_epochs
                >>> }

        Returns: 
            the named scores, recall, precision, f1

        """
        raise NotImplementedError

        return {}

    def do_save(self, predicts: List, stage: str, list_batch_outputs: List[Dict], origin_data: pd.DataFrame, rt_config: Dict, save_condition: bool=False):
        """save the predict when save_condition==True

        Args:
            predicts: list of predicts
            stage: train/test/etc.
            list_batch_outputs: a list of outputs
            origin_data: the origin pd.DataFrame data, there are some data not be able to convert to tensor
            rt_config:
                >>> current status
                >>> {
                >>>     "current_step": self.global_step,
                >>>     "current_epoch": self.current_epoch,
                >>>     "total_steps": self.num_training_steps,
                >>>     "total_epochs": self.num_training_epochs
                >>> }
            save_condition: True for save, False for depend on rt_config

        Returns: 
            None

        """
        if self.config.start_save_epoch == -1 or self.config.start_save_step == -1:
            self.config.start_save_step = rt_config.get('total_steps', 0) - 1
            self.config.start_save_epoch = rt_config.get('total_epochs', 0) - 1
        if not save_condition and (rt_config['current_step']>=self.config.start_save_step or rt_config['current_epoch']>=self.config.start_save_epoch):
            save_condition = True
        if save_condition:
            save_path = os.path.join(self.config.save_root_path, self.config.save_path.get(stage, ''))
            if "current_step" in rt_config:
                save_file = os.path.join(save_path, f"step_{str(rt_config['current_step'])}_predict.json")
            else:
                save_file = os.path.join(save_path, 'predict.json')
            logger.info(f"Save the {stage} predict data at {save_file}")
            with open(save_file, 'w') as f:
                json.dump(predicts, f, indent=4, ensure_ascii=False)



# Copyright 2021 cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hjson
import pandas as pd
from typing import Union, Dict
from dlk.data.postprocessors import postprocessor_register, postprocessor_config_register, IPostProcessor, IPostProcessorConfig
from dlk.utils.config import ConfigTool
import torch

@postprocessor_config_register('identity')
class IdentityPostProcessorConfig(IPostProcessorConfig):
    """docstring for IdentityPostProcessorConfig
    """

    def __init__(self, config: Dict):
        super(IdentityPostProcessorConfig, self).__init__(config)


@postprocessor_register('identity')
class IdentityPostProcessor(IPostProcessor):
    """docstring for DataSet"""
    def __init__(self, config: IdentityPostProcessorConfig):
        super(IdentityPostProcessor, self).__init__(config)

    def process(self, stage, outputs, origin_data)->Dict:
        """do nothing except gather the loss
        """
        if 'loss' in outputs:
            return {self.loss_name_map(stage): torch.mean(outputs['loss'])}
        return {}



# Copyright 2021 cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict
from . import scheduler_register, scheduler_config_register, BaseScheduler, BaseSchedulerConfig
from torch.optim.lr_scheduler import LambdaLR
from dlk.utils.logger import Logger
import torch.optim as optim
logger = Logger.get_logger()


@scheduler_config_register("rec_decay")
class RecDecayScheduleConfig(BaseSchedulerConfig):
    """Config for RecDecaySchedule

    Config Example:
        >>> {
        >>>     "config": {
        >>>         "decay": 0.05, # epoch decay
        >>>     },
        >>>     "_name": "rec_decay",
        >>> }

    the lr=lr*1/(1+decay)
    """
    def __init__(self, config: Dict):
        super(RecDecayScheduleConfig, self).__init__(config)
        config = config['config']
        self.decay = config["decay"]
        self.post_check(config, used=[
            "decay",
        ])


@scheduler_register("rec_decay")
class RecDecaySchedule(BaseScheduler):
    """lr=lr*1/(1+decay)
    """
    def __init__(self, optimizer: optim.Optimizer, config: RecDecayScheduleConfig):
        super(RecDecaySchedule, self).__init__()
        self.config = config
        self.optimizer = optimizer

    def get_scheduler(self):
        """return the initialized rec_decay scheduler

        lr=lr*1/(1+decay)

        Returns: 
            Schedule

        """
        num_training_steps = self.config.num_training_steps
        epoch_training_steps = self.config.epoch_training_steps
        decay = self.config.decay
        logger.warning(f"The calculated Total Traning Num is {num_training_steps}, the Epoch training Steps is {epoch_training_steps}. Please check it carefully.")

        def lr_lambda(current_step: int):
            cur_epoch = (current_step+1)//epoch_training_steps if epoch_training_steps!=0 else 0
            # return 1/(1+decay*cur_epoch)
            return 1/((1+decay)**cur_epoch)
        return LambdaLR(self.optimizer, lr_lambda, last_epoch=1)



# Copyright 2021 cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import setup, find_packages
from dlk import __version__
import os


def package_files(directory):
    paths = []
    for (path, directories, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join('..', path, filename))
    return paths

configures = package_files('dlk/configures')
with open('README.md', encoding='utf-8') as f:
    readme = f.read()

with open('LICENSE', encoding='utf-8') as f:
    license = f.read()

with open('requirements.txt', encoding='utf-8') as f:
    requirements = f.read()

pkgs = [p for p in find_packages() if p.startswith('dlk')]

setup(
    name='dlk',
    version=__version__,
    url='https://github.com/cstsunfu/dlk',
    description='dlk: Deep Learning Kit',
    long_description=readme,
    long_description_content_type='text/markdown',
    license="Apache Software License",
    author='cstsunfu',
    author_email='cstsunfu@gmail.com',
    python_requires='>=3.7',
    package_data={"": configures},
    include_package_data=True,
    packages=pkgs,
    install_requires=requirements.strip().split('\n'),
)



# Copyright 2021 cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hjson
import os
import copy
from typing import Callable, List, Dict, Union
from dlk.utils import register
from dlk.utils.register import Register
from dlk.utils.config import ConfigTool
from dlk.utils.logger import Logger
from dlk.utils.get_root import get_root
from dlk.utils.register import Register
from dlk.core import (
    embedding_config_register,
    embedding_register,
    callback_config_register,
    callback_register,
    decoder_config_register,
    decoder_register,
    encoder_config_register,
    encoder_register,
    imodel_config_register,
    imodel_register,
    initmethod_config_register,
    initmethod_register,
    model_config_register,
    model_register,
    scheduler_config_register,
    loss_config_register,
    loss_register,
    module_config_register,
    module_register,
    optimizer_config_register,
    optimizer_register,
    scheduler_register
)
from dlk.data import (
    datamodule_config_register,
    datamodule_register,
    postprocessor_config_register,
    postprocessor_register,
    processor_config_register,
    processor_register,
    subprocessor_config_register,
    subprocessor_register
)
from dlk.managers import (
    manager_config_register, 
    manager_register
)
import json

logger = Logger.get_logger()

config_parser_register = Register("Config parser register")


class LinkUnionTool(object):
    """Assisting tool for parsering the "_link" of config. All the function named the top level has high priority than low level

    This class is mostly for resolve the confilicts of the low and high level register links.
    """
    def __init__(self):
        self.link_union = {}

    def find(self, key: str):
        """find the root of the key

        Args:
            key: a token

        Returns: 
            the root of the key

        """

        if key not in self.link_union:
            return None

        if self.link_union[key] != key:
            return self.find(self.link_union[key])
        return self.link_union[key]

    def low_level_union(self, link_from: str, link_to: str):
        """union the low level link_from->link_to pair

        On the basis of the high-level links, this function is used to regist low-level link
        If `link-from` and `link-to` were all not appeared at before, they will be directly registed.
        If only one of the `link-from` and `link-to` appeared, the value of the `link-from` and `link-to` will be overwritten by the corresponding value of the upper level,
        If both `link-from` and `link-to` appeared at before, and if they linked the same value, we will do nothing, otherwise `RAISE AN ERROR`

        Args:
            link_from: the link-from key
            link_to: the link-to key

        Returns: 
            None

        """
        if self.find(link_from) and self.find(link_to):  # all has been linked
            if self.find(link_from)!= self.find(link_to):
                raise PermissionError(f"The  {link_from} and {link_to} has been linked to different values, but now you want to link them together.")
            elif self.find(link_from) == link_to:
                logger.warning(f"High level config has the link '{link_to} -> {link_from}', and the low level reversed link '{link_from} -> {link_to}' is been ignored.")
            else:
                return
        elif self.find(link_to): # only link_to has been linked
            logger.warning(f"Parameter '{link_to}' has been linked in high level config, the link '{link_from} -> {link_to}' is invalid, and the real link is been reversed as '{link_to} -> {link_from}'.")
            self.link_union[link_from] = self.find(link_to)
        elif self.find(link_from): # only link_from has been linked
            self.link_union[link_to] = self.find(link_from)
        else:
            self.link_union[link_from] = link_from
            self.link_union[link_to] = link_from

    def top_level_union(self, link_from: str, link_to: str):
        """union the top level link_from->link_to pair

        Register the 'link'(`link-from` -> `link-to`) in the same(top) level config should be merged using `top_level_union`
        Parameters are not allowed to be assigned repeatedly (the same parameter cannot appear more than once in the `link-to` position, otherwise it will cause ambiguity.)

        Args:
            link_from: the link-from key
            link_to: the link-to key

        Returns: 
            None

        """
        if link_from not in self.link_union:
            self.link_union[link_from] = link_from

        assert link_to not in self.link_union, f"{link_to} is repeated assignment"
        self.link_union[link_to] = self.find(link_from)

    def register_top_links(self, links: Dict):
        """register the top level links, top level means the link_to level config

        Args:
            links: {"from": ["tolist"], "from2": "to2"}

        Returns: 
            self

        """
        for source, target in links.items():
            if isinstance(target, list):
                [self.top_level_union(source, t) for t in target]
            else:
                self.top_level_union(source, target)
        return self

    def register_low_links(self, links: Dict):
        """register the low level links, low level means the base(parant) level config

        Args:
            links: {"link-from": ["list of link-to"], "link-from2": "link-to2"}

        Returns: 
            self

        """
        for source, target in links.items():
            if isinstance(target, list):
                [self.low_level_union(source, t) for t in target]
            else:
                self.low_level_union(source, target)
        return self

    def get_links(self):
        """get the registed links

        Returns: 
            all registed and validation links

        """
        links = {}
        for key in self.link_union:
            root = self.find(key)
            if root == key:
                continue
            if root not in links:
                links[root] = []
            links[root].append(key)
        return links


class BaseConfigParser(object):
    """BaseConfigParser
    The config parser order is: inherit -> search -> link

    If some config is marked to "*@*", this means the para has not default value, you must coverd it(like 'label_nums', etc.).

    """
    def __init__(self, config_file: Union[str, Dict, List], config_base_dir: str="", register: Register=None):
        super(BaseConfigParser, self).__init__()
        if isinstance(config_file, str):
            if config_file == '*@*':
                self.config_file = "*@*"
                return
            try:
                if os.path.isfile(os.path.join(get_root(), config_base_dir, config_file+'.hjson')):
                    self.config_file = self.load_hjson_file(os.path.join(get_root(), config_base_dir, config_file+'.hjson'))
                else:
                    self.config_file = copy.deepcopy(register.get(config_file).default_config)

            except Exception as e:
                logger.error(f"You must provide a configure file at {os.path.join(get_root(), config_base_dir, config_file)} or provide `default_config` as a class variable")
                raise KeyError(e)
        elif isinstance(config_file, Dict):
            self.config_file = config_file
        else:
            raise KeyError('The config file must be str or dict. You provide {}.'.format(config_file))

        self.config_name = self.config_file.pop('_name', "")
        self.search = self.config_file.pop('_search', {})
        base = self.config_file.pop('_base', "")
        self.base_config = {}
        if base:
            self.base_config = self.get_base_config(base)
        if "_focus" in self.config_file:
            self.base_config['_focus'] = self.config_file.pop('_focus')

        # merge base and current config _link
        link_union = LinkUnionTool()
        link_union.register_top_links(self.config_file.pop('_link', {}))
        link_union.register_low_links(self.base_config.pop('_link', {}))
        self.config_file['_link'] = link_union.get_links()

        if self.base_config and self.config_name:
            raise PermissionError("You should put the _name to the leaf config.")
        self.modules = self.config_file


    @classmethod
    def get_base_config(cls, config_name: str)->Dict:
        """get the base config use the config_name

        Args:
            config_name: the config name

        Returns: 
            config of the config_name
        """
        base_config = cls(config_name).parser(parser_link=False)
        if len(base_config)>1:
            raise PermissionError("The base config don't support _search now.")
        if base_config:
            return base_config[0]
        return {}

    @staticmethod
    def config_link_para(link: Dict[str, Union[str, List[str]]]=None, config: Dict=None):
        """inplace link the config[to] = config[source]

        Args:
            link: {link-from:link-to-1, link-from:[link-to-2, link-to-3]}
            config: will linked base config

        Returns: 
            None

        """
        if not link:
            link = {}
        if not config:
            config = {}
        def make_link(source: str, to: str):
            """copy the 'source' config to 'to'
            """
            try:
                source_config = config
                to_config = config
                source_list = source.split('.')
                to_list = to.split('.')
                for s in source_list[:-1]:
                    if isinstance(source_config, list):
                        assert (s[0]=='-' and str.isdigit(s[1:])) or str.isdigit(s), "list index must be int"
                        s = int(s)
                    source_config = source_config[s]
                for t in to_list[:-1]:
                    if isinstance(to_config, list):
                        assert (t[0]=='-' and str.isdigit(t[1:])) or str.isdigit(t), "list index must be int"
                        t = int(t)
                    to_config = to_config[t]
                if isinstance(to_config, list):
                    assert (to_list[-1][0]=='-' and str.isdigit(to_list[-1][1:])) or str.isdigit(to_list[-1]), "list index must be int"
                    to_list[-1] = int(to_list[-1])

                if isinstance(source_config, list):
                    assert (source_list[-1][0]=='-' and str.isdigit(source_list[-1][1:])) or str.isdigit(source_list[-1]), "list index must be int"
                    source_list[-1] = int(source_list[-1])

                to_config[to_list[-1]] = source_config[source_list[-1]]
            except Exception as e:
                logger.error(f"Can not link from '{source}' to '{to}'")
                raise e

        if not link:
            return
        for (source, to) in link.items():
            if isinstance(to, List):
                for sub_to in to:
                    make_link(source, sub_to)
            else:
                make_link(source, to)

    @classmethod
    def collect_link(cls, config, trace: List=None, all_level_links: Dict=None, level=0):
        """collect move all links in config to top

        only do in the top level of config, collect all level links and return the links with level

        Args:
            config: 
                >>> {
                >>>     "arg1": {
                >>>         "arg11": 2
                >>>         "arg12": 3
                >>>         "_link": {"arg11": "arg12"}
                >>>     }
                >>> }
            all_level_links: TODO
            level: TODO

        Returns: 
            >>> {
            >>>     "arg1": {
            >>>         "arg11": 2
            >>>         "arg12": 3
            >>>     }
            >>>     "_link": {"arg1.arg11": "arg1.arg12"}
            >>> }

        """
        if not trace:
            trace = []
        if not all_level_links:
            all_level_links = {}
            
        if level not in all_level_links:
            all_level_links[level] = {}
        trace_str = ".".join(trace)
        if trace_str:
            # except the top level, all add the '.' suffix
            trace_str = trace_str + '.'
        def add_trace(origin_link: Dict)->Dict:
            """add the root of the config to current config trace to current level para of links

            Args:
                origin_link: which is not added the trace(root to cur node)

            Returns: added trace link

            """
            added_trace_link = {}
            for source, target in origin_link.items():
                if isinstance(target, list):
                    target = [trace_str+t for t in target]
                else:
                    target = trace_str + target
                source = trace_str + source
                added_trace_link[source] = target
            return added_trace_link

        if "_link" in config:
            # all_level_links[level] = add_trace(config['_link'])
            all_level_links[level].update(add_trace(config.pop('_link')))

        for submodule_name, submodule_config in config.items():
            if isinstance(submodule_config, dict):
                cls.collect_link(submodule_config, trace + [submodule_name], all_level_links, level+1)
        return all_level_links

    def parser_with_check(self, parser_link=True)->List[Dict]:
        """parser the config and check the config is valid

        Args:
            parser_link: whether parser the links

        Returns: all valided configs

        """
        configs = self.parser(parser_link)
        self.check_config(configs)
        return configs

    def parser(self, parser_link=True) -> List:
        """parser the config

        Args:
            parser_link: whether parser the links

        Returns: all valided configs

        """
        if self.config_file == '*@*':
            return ['*@*']

        # parser submodules get submodules config
        modules_config = self.map_to_submodule(self.modules, self.get_kind_module_base_config)

        # expand all submodules to combine a set of module configs
        possible_config_list = self.get_named_list_cartesian_prod(modules_config)

        # using specifical module config to update base_config
        if possible_config_list:
            possible_config_list = [ConfigTool.do_update_config(self.base_config, possible_config) for possible_config in possible_config_list]
        else:
            possible_config_list = [self.base_config]

        # flat all search paras
        possible_config_list_list = [self.flat_search(self.search, possible_config) for possible_config in possible_config_list]

        all_possible_config_list = []
        for possible_config_list in possible_config_list_list:
            all_possible_config_list.extend(possible_config_list)

        # link paras
        if parser_link:
            for possible_config in all_possible_config_list:
                all_level_links = self.collect_link(possible_config)
                link_union = LinkUnionTool()
                for i in range(len(all_level_links)):
                    cur_level_links = all_level_links[i]
                    link_union.register_low_links(cur_level_links)

                self.config_link_para(link_union.get_links(), possible_config)

        return_list = []
        for possible_config in all_possible_config_list:
            config = copy.deepcopy(possible_config)
            if self.config_name:
                config['_name'] = self.config_name
            return_list.append(config)

        if self.is_rep_config(return_list):
            logger.warning(f"The Configures is Repeated, Please Check The Configures Carefully.")
            for i, config in enumerate(return_list):
                logger.info(f"The {i}th Configure is:")
                logger.info(json.dumps(config, indent=2, ensure_ascii=False))
            raise ValueError('REPEAT CONFIG')
        return return_list

    def get_kind_module_base_config(self, abstract_config: Union[dict, str], kind_module: str="") -> List[dict]:
        """get the whole config of 'kind_module' by given abstract_config

        Args:
            abstract_config: will expanded config
            kind_module: the module kind, like 'embedding', 'subprocessor', which registed in config_parser_register

        Returns: parserd config (whole config) of abstract_config

        """
        return config_parser_register.get(kind_module)(abstract_config).parser(parser_link=False)

    def map_to_submodule(self, config: dict, map_fun: Callable) -> Dict:
        """map the map_fun to all submodules in config

        use the map_fun to process all the modules

        Args:
            config: a dict of submodules, the key is the module kind wich registed in config_parser_register
            map_fun: use the map_fun process the submodule

        Returns: TODO

        """
        modules_config = {}
        for kind_module in config:
            modules_config[kind_module] = map_fun(config[kind_module], kind_module)
        return modules_config

    def load_hjson_file(self, file_path: str) -> Dict:
        """load hjson file from file_path and return a Dict

        Args:
            file_path: the file path

        Returns: loaded dict

        """
        json_file = hjson.load(open(file_path), object_pairs_hook=dict)
        return json_file

    @classmethod
    def flat_search(cls, search, config: dict) -> List[dict]:
        """flat all the _search paras to list

        support recursive parser _search now, this means you can add _search/_link/_base paras in _search paras
        but you should only search currently level paras

        Args:
            search: search paras, {"para1": [1,2,3], 'para2': 'list(range(10))'}
            config: base config

        Returns: list of possible config

        """
        result = []
        module_search_para = search
        if not module_search_para:
            result.append(config)
        else:
            search_para_list = cls.get_named_list_cartesian_prod(module_search_para)
            for search_para in search_para_list:
                base_config = copy.deepcopy(config)
                base_config.update(search_para)
                extend_config = cls(base_config).parser(parser_link=False)
                result.extend(extend_config)

        return result

    def get_cartesian_prod(self, list_of_list_of_dict: List[List[Dict]]) -> List[List[Dict]]:
        """get catesian prod from two lists

        Args:
            list_of_list_of_dict: [[config_a1, config_a2], [config_b1, config_b2]]

        Returns: 
            [[config_a1, config_b1], [config_a1, config_b2], [config_a2, config_b1], [config_a2, config_b2]]

        """
        if len(list_of_list_of_dict) <= 1:
            return [copy.deepcopy(dic) for dic in list_of_list_of_dict]
        cur_result = list_of_list_of_dict[0]
        reserve_result = self.get_cartesian_prod(list_of_list_of_dict[1:])
        result = []
        for cur_config in cur_result:
            for reserve in reserve_result:
                result.append([copy.deepcopy(cur_config)]+copy.deepcopy(reserve))
        return result

    @staticmethod
    def check_config(configs: Union[Dict, List[Dict]]) -> None:
        """check all config is valid.

        check all "*@*" is replaced to correct value.
        Args:
            configs: TODO

        Returns: 
            None

        Raises:
            ValueError

        """
        def _check(config):
            """check the "*@*" is in config or not
            """
            for key in config:
                if isinstance(config[key], dict):
                    _check(config[key])
                if config[key] == '*@*':
                    raise ValueError(f'In Config: \n {json.dumps(config, indent=4, ensure_ascii=False)}\n The must be provided key "{key}" marked with "*@*" is not provided.')

        if isinstance(configs, list):
            for config in configs:
                _check(config)
        else:
            _check(configs)

    @staticmethod
    def get_named_list_cartesian_prod(dict_of_list: Dict[str, List]=None) -> List[Dict]:
        """get catesian prod from named lists

        Args:
            dict_of_list: {'name1': [1,2,3], 'name2': "list(range(1, 4))"}

        Returns: 
            [{'name1': 1, 'name2': 1}, {'name1': 1, 'name2': 2}, {'name1': 1, 'name2': 3}, ...]

        """
        if not dict_of_list:
            dict_of_list = {}
        if len(dict_of_list) == 0:
            return []
        dict_of_list = copy.deepcopy(dict_of_list)
        cur_name, cur_paras  = dict_of_list.popitem()
        cur_para_search_list = []
        if isinstance(cur_paras, str):
            cur_paras = eval(cur_paras)
        assert isinstance(cur_paras, list), f"The search candidates must be list, but you provide {cur_paras}({type(cur_paras)})"
        for para in cur_paras:
            cur_para_search_list.append({cur_name: para})
        if len(dict_of_list) == 0:
            return cur_para_search_list
        reserve_para_list = BaseConfigParser.get_named_list_cartesian_prod(dict_of_list)
        all_config_list = []
        for cur_config in cur_para_search_list:
            for reserve_config in reserve_para_list:
                _cur_config = copy.deepcopy(cur_config)
                _cur_config.update(copy.deepcopy(reserve_config))
                all_config_list.append(_cur_config)
        return all_config_list

    def is_rep_config(self, list_of_dict: List[dict]) -> bool:
        """check is there a repeat config in list

        Args:
            list_of_dict: a list of dict

        Returns:
            has repeat or not

        """
        # using json.dumps + sort_keys to guarantee the same dict to the same string represatation
        list_of_str = [json.dumps(dic, sort_keys=True, ensure_ascii=False) for dic in list_of_dict]
        if len(list_of_dict) == len(set(list_of_str)):
            return False
        else:
            return True


@config_parser_register('config')
class ConfigConfigParser(BaseConfigParser):
    """ConfigConfigParser"""
    def __init__(self, config_file):
        super(ConfigConfigParser, self).__init__(config_file, config_base_dir='NONEPATH')
        if self.base_config:
            raise AttributeError('The paras config do not support _base.')

        if self.config_name:
            raise AttributeError('The paras config do not support _name.')

    def parser(self, parser_link=True):
        """parser the config

        config support _search and _link

        Args:
            parser_link: whether parser the links

        Returns:
            all valided configs

        """
        config_list = self.flat_search(self.search, self.modules)
        # link paras
        if parser_link:
            for possible_config in config_list:
                all_level_links = self.collect_link(possible_config)
                link_union = LinkUnionTool()
                for i in range(len(all_level_links)):
                    cur_level_links = all_level_links[i]
                    link_union.register_low_links(cur_level_links)
                self.config_link_para(all_level_links, possible_config)

        return config_list


@config_parser_register('_link')
class LinkConfigParser(object):
    """LinkConfigParser"""
    def __init__(self, config_file):
        self.config = config_file
        assert isinstance(self.config, dict), f"The '_link' must be a dict, but you provide '{self.config}'"
    def parser(self, parser_link=False):
        """parser the config

        config support _search and _link

        Args:
            parser_link: must be false

        Returns:
            all valided configs

        """
        assert parser_link is False, f"The parser_link para must be False when parser the _link"
        return [self.config]


module_dir_map = {
    "task": "dlk/configures/tasks",
    "manager": "dlk/configures/managers",
    "callback": "dlk/configures/core/callbacks",
    "datamodule": "dlk/configures/data/datamodules",
    "imodel": "dlk/configures/core/imodels",
    "model": "dlk/configures/core/models",
    "optimizer": "dlk/configures/core/optimizers",
    "scheduler": "dlk/configures/core/schedulers",
    "initmethod": "dlk/configures/core/initmethods",
    "loss": "dlk/configures/core/losses",
    "encoder": "dlk/configures/core/layers/encoders",
    "decoder": "dlk/configures/core/layers/decoders",
    "embedding": "dlk/configures/core/layers/embeddings",
    "module": "dlk/configures/core/modules",
    "processor": "dlk/configures/data/processors",
    "subprocessor": "dlk/configures/data/subprocessors",
    "postprocessor": "dlk/configures/data/postprocessors",
}


@config_parser_register('task')
class TaskConfigParser(BaseConfigParser):
    """docstring for TaskConfigParser"""
    def __init__(self, config_file):
        super(TaskConfigParser, self).__init__(config_file, config_base_dir=module_dir_map['task'])


@config_parser_register('root')
class RootConfigParser(BaseConfigParser):
    """docstring for RootConfigParser"""
    def __init__(self, config_file):
        super(RootConfigParser, self).__init__(config_file, config_base_dir='')


@config_parser_register('manager')
class ManagerConfigParser(BaseConfigParser):
    """docstring for ManagerConfigParser"""
    def __init__(self, config_file):
        super(ManagerConfigParser, self).__init__(config_file, config_base_dir=module_dir_map['manager'], register=manager_config_register)


@config_parser_register('callback')
class CallbackConfigParser(BaseConfigParser):
    """docstring for CallbackConfigParser"""
    def __init__(self, config_file):
        super(CallbackConfigParser, self).__init__(config_file, config_base_dir=module_dir_map['callback'], register=callback_config_register)


@config_parser_register('datamodule')
class DatamoduleConfigParser(BaseConfigParser):
    """docstring for DatamoduleConfigParser"""
    def __init__(self, config_file):
        super(DatamoduleConfigParser, self).__init__(config_file, config_base_dir=module_dir_map['datamodule'], register=datamodule_config_register)


@config_parser_register('imodel')
class IModelConfigParser(BaseConfigParser):
    """docstring for IModelConfigParser"""
    def __init__(self, config_file):
        super(IModelConfigParser, self).__init__(config_file, config_base_dir=module_dir_map['imodel'], register=imodel_config_register)


@config_parser_register('model')
class ModelConfigParser(BaseConfigParser):
    """docstring for ModelConfigParser"""
    def __init__(self, config_file):
        super(ModelConfigParser, self).__init__(config_file, config_base_dir=module_dir_map['model'], register=model_config_register)


@config_parser_register('optimizer')
class OptimizerConfigParser(BaseConfigParser):
    """docstring for OptimizerConfigParser"""
    def __init__(self, config_file):
        super(OptimizerConfigParser, self).__init__(config_file, config_base_dir=module_dir_map['optimizer'], register=optimizer_config_register)

@config_parser_register('scheduler')
class ScheduleConfigParser(BaseConfigParser):
    """docstring for ScheduleConfigParser"""
    def __init__(self, config_file):
        super(ScheduleConfigParser, self).__init__(config_file, config_base_dir=module_dir_map['scheduler'], register=scheduler_config_register)

@config_parser_register('initmethod')
class InitMethodConfigParser(BaseConfigParser):
    """docstring for InitMethodConfigParser"""
    def __init__(self, config_file):
        super(InitMethodConfigParser, self).__init__(config_file, config_base_dir=module_dir_map['initmethod'], register=initmethod_config_register)


@config_parser_register('loss')
class LossConfigParser(BaseConfigParser):
    """docstring for LossConfigParser"""
    def __init__(self, config_file):
        super(LossConfigParser, self).__init__(config_file, config_base_dir=module_dir_map['loss'], register=loss_config_register)


@config_parser_register('encoder')
class EncoderConfigParser(BaseConfigParser):
    """docstring for EncoderConfigParser"""
    def __init__(self, config_file):
        super(EncoderConfigParser, self).__init__(config_file, config_base_dir=module_dir_map['encoder'], register=encoder_config_register)


@config_parser_register('decoder')
class DecoderConfigParser(BaseConfigParser):
    """docstring for DecoderConfigParser"""
    def __init__(self, config_file):
        super(DecoderConfigParser, self).__init__(config_file, config_base_dir=module_dir_map['decoder'], register=decoder_config_register)


@config_parser_register('embedding')
class EmbeddingConfigParser(BaseConfigParser):
    """docstring for EmbeddingConfigParser"""
    def __init__(self, config_file):
        super(EmbeddingConfigParser, self).__init__(config_file, config_base_dir=module_dir_map['embedding'], register=embedding_config_register)


@config_parser_register('module')
class ModuleConfigParser(BaseConfigParser):
    """docstring for ModuleConfigParser"""
    def __init__(self, config_file):
        super(ModuleConfigParser, self).__init__(config_file, config_base_dir=module_dir_map['module'], register=module_config_register)

@config_parser_register('processor')
class ProcessorConfigParser(BaseConfigParser):
    """docstring for ProcessorConfigParser"""
    def __init__(self, config_file):
        super(ProcessorConfigParser, self).__init__(config_file, config_base_dir=module_dir_map['processor'], register=processor_config_register)


@config_parser_register('subprocessor')
class SubProcessorConfigParser(BaseConfigParser):
    """docstring for SubProcessorConfigParser"""
    def __init__(self, config_file):
        super(SubProcessorConfigParser, self).__init__(config_file, config_base_dir=module_dir_map['subprocessor'], register=subprocessor_config_register)


@config_parser_register('postprocessor')
class PostProcessorConfigParser(BaseConfigParser):
    """docstring for PostProcessorConfigParser"""
    def __init__(self, config_file):
        super(PostProcessorConfigParser, self).__init__(config_file, config_base_dir=module_dir_map['postprocessor'], register=postprocessor_config_register)




# Copyright 2021 cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict
from . import callback_register, callback_config_register
from pytorch_lightning.callbacks import LearningRateMonitor


@callback_config_register('lr_monitor')
class LearningRateMonitorCallbackConfig(object):
    """Config for LearningRateMonitorCallback

    Config Example:
        >>> {
        >>>     "_name": "lr_monitor",
        >>>     "config": {
        >>>         "logging_interval": null, // set to None to log at individual interval according to the interval key of each scheduler. other value : step, epoch
        >>>         "log_momentum": true, // log momentum or not
        >>>     }
        >>> }
    """
    def __init__(self, config: Dict):
        config = config['config']
        self.logging_interval = config["logging_interval"]
        self.log_momentum = config["log_momentum"]


@callback_register('lr_monitor')
class LearningRateMonitorCallback(object):
    """Monitor the learning rate
    """

    def __init__(self, config: LearningRateMonitorCallbackConfig):
        super().__init__()
        self.config = config

    def __call__(self, rt_config: Dict)->LearningRateMonitor:
        """return LearningRateMonitor object

        Args:
            rt_config: runtime config, include save_dir, and the checkpoint path name

        Returns: 
            LearningRateMonitor object

        """
        return LearningRateMonitor(**self.config.__dict__)



# Copyright 2021 cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch.nn as nn
from dlk.utils.config import BaseConfig
import torch
from typing import Dict, List, Collection
from . import module_register, module_config_register, Module

@module_config_register("conv1d")
class Conv1dConfig(BaseConfig):
    """Config for Conv1d

    Config Example:
        >>> {
        >>>     "config": {
        >>>         "in_channels": "*@*",
        >>>         "out_channels": "*@*",
        >>>         "dropout": 0.0,
        >>>         "kernel_sizes": [3],
        >>>     },
        >>>     "_name": "conv1d",
        >>> }
    """
    def __init__(self, config: Dict):
        super(Conv1dConfig, self).__init__(config)
        config = config['config']
        self.kernel_sizes = config['kernel_sizes']
        out_channels = config['out_channels']
        assert all(k % 2 == 1 for k in self.kernel_sizes), 'the kernel sizes must be odd'
        assert out_channels % len(self.kernel_sizes) == 0, 'out channels must be dividable by kernels'
        self.in_channels = config['in_channels']
        self.out_channels = out_channels // len(self.kernel_sizes)
        self.dropout = config['dropout']
        self.post_check(config, used=[
            "in_channels",
            "out_channels",
            "dropout",
            "kernel_sizes",
        ])

@module_register("conv1d")
class Conv1d(Module):
    """Conv for 1d input
    """
    def __init__(self, config: Conv1dConfig):
        super().__init__()
        convs = []
        for kernel_size in config.kernel_sizes:
            conv = nn.Conv1d(config.in_channels, config.out_channels, kernel_size,
                             padding=(kernel_size - 1) // 2)
            convs.append(nn.Sequential(conv, nn.GELU()))
        self.convs = nn.ModuleList(convs)
        self.dropout = nn.Dropout(p=float(config.dropout))

    def forward(self, x: torch.Tensor):
        """do forward on a mini batch

        Args:
            batch: a mini batch inputs

        Returns: 
            conv result the shape is the same as input

        """
        return self.dropout(torch.cat([conv(x) for conv in self.convs], dim=-1))



# Copyright 2021 cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import model_register, model_config_register
from typing import Dict, List
from dlk.utils.config import BaseConfig, ConfigTool
from dlk.core.base_module import BaseModel
import torch
from dlk.core.layers.embeddings import embedding_config_register, embedding_register
from dlk.core.initmethods import initmethod_config_register, initmethod_register
from dlk.core.layers.encoders import encoder_config_register, encoder_register
from dlk.core.layers.decoders import decoder_config_register, decoder_register


@model_config_register('basic')
class BasicModelConfig(BaseConfig):
    """Config for BasicModel

    Config Example:
        >>> {
        >>>     embedding: {
        >>>         _base: "static"
        >>>         config: {
        >>>             embedding_file: "*@*", //the embedding file, must be saved as numpy array by pickle
        >>>             embedding_dim: "*@*",
        >>>             //if the embedding_file is a dict, you should provide the dict trace to embedding
        >>>             embedding_trace: ".", //default the file itself is the embedding
        >>>             /*embedding_trace: "embedding", //this means the <embedding = pickle.load(embedding_file)["embedding"]>*/
        >>>             /*embedding_trace: "meta.embedding", //this means the <embedding = pickle.load(embedding_file)['meta']["embedding"]>*/
        >>>             freeze: false, // is freeze
        >>>             dropout: 0, //dropout rate
        >>>             output_map: {},
        >>>         },
        >>>     },
        >>>     decoder: {
        >>>         _base: "linear",
        >>>         config: {
        >>>             input_size: "*@*",
        >>>             output_size: "*@*",
        >>>             pool: null,
        >>>             dropout: "*@*", //the decoder output no need dropout
        >>>             output_map: {}
        >>>         },
        >>>     },
        >>>     encoder: {
        >>>         _base: "lstm",
        >>>         config: {
        >>>             output_map: {},
        >>>             hidden_size: "*@*",
        >>>             input_size: *@*,
        >>>             output_size: "*@*",
        >>>             num_layers: 1,
        >>>             dropout: "*@*", // dropout between layers
        >>>         },
        >>>     },
        >>>     "initmethod": {
        >>>         "_base": "range_norm"
        >>>     },
        >>>     "config": {
        >>>         "embedding_dim": "*@*",
        >>>         "dropout": "*@*",
        >>>         "embedding_file": "*@*",
        >>>         "embedding_trace": "token_embedding",
        >>>     },
        >>>     _link: {
        >>>         "config.embedding_dim": ["embedding.config.embedding_dim",
        >>>                                  "encoder.config.input_size",
        >>>                                  "encoder.config.output_size",
        >>>                                  "encoder.config.hidden_size",
        >>>                                  "decoder.config.output_size",
        >>>                                  "decoder.config.input_size"
        >>>                                 ],
        >>>         "config.dropout": ["encoder.config.dropout", "decoder.config.dropout", "embedding.config.dropout"],
        >>>         "config.embedding_file": ['embedding.config.embedding_file'],
        >>>         "config.embedding_trace": ['embedding.config.embedding_trace']
        >>>     }
        >>>     _name: "basic"
        >>> }
    """
    def __init__(self, config):
        super(BasicModelConfig, self).__init__(config)

        self.embedding, self.embedding_config = self.get_embedding(config["embedding"])
        self.encoder, self.encoder_config = self.get_encoder(config["encoder"])
        self.decoder, self.decoder_config = self.get_decoder(config["decoder"])
        self.init_method, self.init_method_config = self.get_init_method(config["initmethod"])
        self.config = config['config']

    def get_embedding(self, config: Dict):
        """return the Embedding and EmbeddingConfig

        Args:
            config: the embedding config

        Returns: 
            Embedding, EmbeddingConfig

        """
        return ConfigTool.get_leaf_module(embedding_register, embedding_config_register, "embedding", config)

    def get_init_method(self, config: Dict):
        """return the InitMethod and InitMethodConfig

        Args:
            config: the init method config

        Returns: 
            InitMethod, InitMethodConfig

        """
        return ConfigTool.get_leaf_module(initmethod_register, initmethod_config_register, "init method", config)

    def get_encoder(self, config: Dict):
        """return the Encoder and EncoderConfig

        Args:
            config: the encoder config

        Returns: 
            Encoder, EncoderConfig

        """
        return ConfigTool.get_leaf_module(encoder_register, encoder_config_register, "encoder", config)

    def get_decoder(self, config):
        """return the Decoder and DecoderConfig

        Args:
            config: the decoder config

        Returns: 
            Decoder, DecoderConfig

        """
        return ConfigTool.get_leaf_module(decoder_register, decoder_config_register, "decoder", config)


@model_register('basic')
class BasicModel(BaseModel):
    """Basic & General Model
    """

    def __init__(self, config: BasicModelConfig, checkpoint):
        super().__init__()

        self.embedding = config.embedding(config.embedding_config)
        self.encoder = config.encoder(config.encoder_config)
        self.decoder = config.decoder(config.decoder_config)

        if not checkpoint:
            init_method = config.init_method(config.init_method_config)
            self.embedding.init_weight(init_method)
            self.encoder.init_weight(init_method)
            self.decoder.init_weight(init_method)

        self.config = config.config
        self._provided_keys = self.config.get("provided_keys", [])

    def provide_keys(self)->List[str]:
        """return all keys of the dict of the model returned

        This method may no use, so we will remove this.

        Returns: all keys
        """
        return self.decoder.provided_keys()

    def check_keys_are_provided(self, provide: List[str]=[])->None:
        """check this all the submodules required key are provided

        Returns: None

        Raises: PermissionError

        """
        self._provided_keys = self._provided_keys + provide
        self.embedding.check_keys_are_provided(self._provided_keys)
        self.encoder.check_keys_are_provided(self.embedding.provide_keys())
        self.decoder.check_keys_are_provided(self.encoder.provide_keys())

    def forward(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        """do forward on a mini batch

        Args:
            batch: a mini batch inputs

        Returns: the outputs

        """
        embedding_outputs = self.embedding(inputs)
        encode_outputs = self.encoder(embedding_outputs)
        decode_outputs = self.decoder(encode_outputs)
        return decode_outputs

    def predict_step(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        """do predict for one batch

        Args:
            inputs: one mini-batch inputs

        Returns: the predicts outputs

        """
        embedding_outputs = self.embedding.predict_step(inputs)
        encode_outputs = self.encoder.predict_step(embedding_outputs)
        decode_outputs = self.decoder.predict_step(encode_outputs)
        return decode_outputs

    def training_step(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        """do training for one batch

        Args:
            inputs: one mini-batch inputs

        Returns: the training outputs

        """
        embedding_outputs = self.embedding.training_step(inputs)
        encode_outputs = self.encoder.training_step(embedding_outputs)
        decode_outputs = self.decoder.training_step(encode_outputs)
        return decode_outputs

    def validation_step(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        """do validation for one batch

        Args:
            inputs: one mini-batch inputs

        Returns: the validation outputs

        """
        embedding_outputs = self.embedding.validation_step(inputs)
        encode_outputs = self.encoder.validation_step(embedding_outputs)
        decode_outputs = self.decoder.validation_step(encode_outputs)
        return decode_outputs

    def test_step(self, inputs: Dict[str, torch.Tensor])->Dict[str, torch.Tensor]:
        """do test for one batch

        Args:
            inputs: one mini-batch inputs

        Returns: the test outputs

        """
        embedding_outputs = self.embedding.test_step(inputs)
        encode_outputs = self.encoder.test_step(embedding_outputs)
        decode_outputs = self.decoder.test_step(encode_outputs)
        return decode_outputs



# Copyright 2021 cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dlk.utils.parser import BaseConfigParser
import json
import hjson
from typing import Dict, Union, Any
from dlk.data.processors import processor_config_register, processor_register
from dlk.utils.io import open


class Processor(object):
    """Processor"""
    def __init__(self, config: Union[str, Dict]):
        super(Processor, self).__init__()
        if not isinstance(config, dict):
            with open(config) as f:
                config = hjson.load(f, object_pairs_hook=dict)
            config = BaseConfigParser(config).parser_with_check()
            assert len(config) == 1, f"Currently we didn't support search for Processor, if you require this feature please create an issue to describe the reason details."
            self.config = config[0]
        self.config = self.config['processor']
        
    def fit(self, data: Dict[str, Any], stage='train'):
        """Process the data and return the processed data

        Args:
            data: {"train": .., 'valid': ..}
            stage: "train"/ 'predict', etc.

        Returns: 
            processed data

        """
        
        processor = processor_register.get(self.config.get('_name'))(stage=stage, config=processor_config_register.get(self.config.get('_name'))(stage=stage, config=self.config))
        return processor.process(data)



# Copyright 2021 cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict
import torch.nn as nn
import torch.optim as optim
from dlk.utils.config import BaseConfig
from . import optimizer_register, optimizer_config_register, BaseOptimizer


@optimizer_config_register("adamw")
class AdamWOptimizerConfig(BaseConfig):
    """Config for AdamWOptimizer

    Config Example:
        >>> {
        >>>     "config": {
        >>>         "lr": 5e-5,
        >>>         "betas": [0.9, 0.999],
        >>>         "eps": 1e-6,
        >>>         "weight_decay": 1e-2,
        >>>         "optimizer_special_groups": {
        >>>             "order": ['decoder', 'bias'], // the group order, if the para is in decoder & is in bias, set to decoder. The order name is set to the group name
        >>>             "bias": {
        >>>                 "config": {
        >>>                     "weight_decay": 0
        >>>                 },
        >>>                 "pattern": ["bias",  "LayerNorm.bias", "LayerNorm.weight"]
        >>>             },
        >>>             "decoder": {
        >>>                 "config": {
        >>>                     "lr": 1e-3
        >>>                 },
        >>>                 "pattern": ["decoder"]
        >>>             },
        >>>         }
        >>>         "name": "default" // default group name
        >>>     },
        >>>     "_name": "adamw",
        >>> }
    """
    def __init__(self, config: Dict):
        super(AdamWOptimizerConfig, self).__init__(config)
        self.config = config['config']
        self.post_check(self.config, used=[
            "lr",
            "betas",
            "eps",
            "weight_decay",
            "optimizer_special_groups",
            "name",
        ])


@optimizer_register("adamw")
class AdamWOptimizer(BaseOptimizer):
    """Wrap for optim.AdamW
    """
    def __init__(self, model: nn.Module, config: AdamWOptimizerConfig):
        super(AdamWOptimizer, self).__init__()
        self.config = config.config
        self.model = model
        self.optimizer = optim.AdamW

    def get_optimizer(self)->optim.AdamW:
        """return the initialized AdamW optimizer

        Returns: 
            AdamW Optimizer

        """
        return self.init_optimizer(optim.AdamW, self.model, self.config)



# Copyright 2021 cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""schedulers"""
import importlib
import os
from dlk.utils.register import Register
from torch.optim import Optimizer
from dlk.utils.config import BaseConfig
from torch.optim.lr_scheduler import LambdaLR
import math


scheduler_config_register = Register("Schedule config register.")
scheduler_register = Register("Schedule register.")

class BaseSchedulerConfig(BaseConfig):
    """interface for Schedule"""
    def __init__(self, config):
        super(BaseSchedulerConfig, self).__init__(config)
        self.num_training_epochs :int = None
        self.epoch_training_steps :int = None
        self.num_training_steps :int = None
        self.last_epoch :int = None

class BaseScheduler(object):
    """interface for Schedule"""

    def get_scheduler(self)->LambdaLR:
        """return the initialized scheduler

        Returns: 
            Schedule

        """
        raise NotImplementedError

    def __call__(self):
        """the same as self.get_scheduler()
        """
        return self.get_scheduler()


def import_schedulers(schedulers_dir, namespace):
    for file in os.listdir(schedulers_dir):
        path = os.path.join(schedulers_dir, file)
        if (
            not file.startswith("_")
            and not file.startswith(".")
            and (file.endswith(".py") or os.path.isdir(path))
        ):
            scheduler_name = file[: file.find(".py")] if file.endswith(".py") else file
            importlib.import_module(namespace + "." + scheduler_name)


# automatically import any Python files in the schedulers directory
schedulers_dir = os.path.dirname(__file__)
import_schedulers(schedulers_dir, "dlk.core.schedulers")



# Copyright 2021 cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hjson
import pandas as pd
from typing import Union, Dict
from dlk.data.processors import IProcessor, processor_config_register, processor_register
from dlk.data.subprocessors import subprocessor_config_register, subprocessor_register
from dlk.utils.config import BaseConfig
from dlk.utils.logger import Logger
from dlk.utils.io import open

logger = Logger.get_logger()


@processor_config_register('basic')
class BasicProcessorConfig(BaseConfig):
    """Config for BasicProcessor

    Config Example:
        >>> {
        >>>     // input should be {"train": train, "valid": valid, ...}, train/valid/test/predict/online etc, should be dataframe and must have a column named "sentence"
        >>>     "_name": "basic@test_text_cls",
        >>>     "config": {
        >>>         "feed_order": ["load", "tokenizer", "token_gather", "label_to_id", "token_embedding", "save"]
        >>>     },
        >>>     "subprocessor@load": {
        >>>         "_base": "load",
        >>>         "config":{
        >>>             "base_dir": "",
        >>>             "predict":{
        >>>                 "meta": "./meta.pkl",
        >>>             },
        >>>             "online": [
        >>>                 "predict", //base predict
        >>>                 {   // special config, update predict, is this case, the config is null, means use all config from "predict", when this is empty dict, you can only set the value to a str "predict", they will get the same result
        >>>                 }
        >>>             ]
        >>>         }
        >>>     },
        >>>     "subprocessor@save": {
        >>>         "_base": "save",
        >>>         "config":{
        >>>             "base_dir": "",
        >>>             "train":{
        >>>                 "processed": "processed_data.pkl", // all data
        >>>                 "meta": {
        >>>                     "meta.pkl": ['label_vocab'] //only for next time use
        >>>                 }
        >>>             },
        >>>             "predict": {
        >>>                 "processed": "processed_data.pkl",
        >>>             }
        >>>         }
        >>>     },
        >>>     "subprocessor@tokenizer":{
        >>>         "_base": "fast_tokenizer",
        >>>         "config": {
        >>>             "train": {
        >>>                 "config_path": "*@*",
        >>>                 "prefix": ""
        >>>                 "data_type": "single", // single or pair, if not provide, will calc by len(process_data)
        >>>                 "process_data": [
        >>>                     ["sentence", { "is_pretokenized": false}],
        >>>                 ],
        >>>                 "post_processor": "default"
        >>>                 "filed_map": { // this is the default value, you can provide other name
        >>>                     "ids": "input_ids",
        >>>                 }, // the tokenizer output(the key) map to the value
        >>>             },
        >>>             "predict": "train",
        >>>             "online": "train"
        >>>         }
        >>>     },
        >>>     "subprocessor@token_gather":{
        >>>         "_base": "token_gather",
        >>>         "config": {
        >>>             "train": { // only train stage using
        >>>                 "data_set": {      // for different stage, this processor will process different part of data
        >>>                     "train": ["train", "valid"]
        >>>                 },
        >>>                 "gather_columns": ["label"], //List of columns. Every cell must be sigle token or list of tokens or set of tokens
        >>>                 "deliver": "label_vocab", // output Vocabulary object (the Vocabulary of labels) name.
        >>>             }
        >>>         }
        >>>     },
        >>>     "subprocessor@label_to_id":{
        >>>         "_base": "token2id",
        >>>         "config": {
        >>>             "train":{ //train、predict、online stage config,  using '&' split all stages
        >>>                 "data_pair": {
        >>>                     "label": "label_id"
        >>>                 },
        >>>                 "data_set": {                   // for different stage, this processor will process different part of data
        >>>                     "train": ['train', 'valid', 'test'],
        >>>                     "predict": ['predict'],
        >>>                     "online": ['online']
        >>>                 },
        >>>                 "vocab": "label_vocab", // usually provided by the "token_gather" module
        >>>             }, //3
        >>>             "predict": "train",
        >>>             "online": "train",
        >>>         }
        >>>     },
        >>>     "subprocessor@token_embedding": {
        >>>         "_base": "token_embedding",
        >>>         "config":{
        >>>             "train": { // only train stage using
        >>>                 "embedding_file": "*@*",
        >>>                 "tokenizer": "*@*", //List of columns. Every cell must be sigle token or list of tokens or set of tokens
        >>>                 "deliver": "token_embedding", // output Vocabulary object (the Vocabulary of labels) name.
        >>>                 "embedding_size": 200,
        >>>             }
        >>>         }
        >>>     },
        >>> }
    """
    def __init__(self, stage, config: Dict):
        super(BasicProcessorConfig, self).__init__(config)
        if isinstance(config, str):
            with open(config) as f:
                config = hjson.load(f, object_pairs_hook=dict)
        self.feed_order = config["config"]['feed_order']
        self.subprocessors = config
        self.stage = stage
        # self.post_check(config['config'], used=['feed_order'])


@processor_register('basic')
class BasicProcessor(IProcessor):
    """Basic and General Processor"""
    def __init__(self, stage: str, config: BasicProcessorConfig):
        super(BasicProcessor, self).__init__()
        self._name = config._name
        self.stage = stage
        self.feed_order = config.feed_order
        assert len(self.feed_order) > 0

        self.subprocessors = {}
        for name in self.feed_order:
            subprocessor_config_dict = config.subprocessors[
                f'subprocessor@{name}']
            logger.info(f"Init '{name}' ....")
            subprocessor_name = subprocessor_config_dict["_name"]
            subprocessor_config = subprocessor_config_register.get(
                subprocessor_name)(stage=self.stage,
                                   config=subprocessor_config_dict)
            subprocessor = subprocessor_register.get(subprocessor_name)(
                stage=self.stage, config=subprocessor_config)
            self.subprocessors[name] = subprocessor

    def process(self, data: Dict) -> Dict:
        """Process entry

        Args:
            data: 
            >>> {
            >>>     "data": {"train": ...},
            >>>     "tokenizer": ..
            >>> }

        Returns: 
            processed data

        """
        logger.info(f"Start Data Processing....")
        for name in self.feed_order:
            if self.stage != 'online':
                logger.info(f"Processing on '{name}' ....")
            data = self.subprocessors[name].process(data)
        logger.info(f"Data Processed.")

        return data



# Copyright cstsunfu. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dlk.utils.vocab import Vocabulary
from dlk.utils.config import BaseConfig, ConfigTool
from typing import Dict, Callable, Set, List
from dlk.data.subprocessors import subprocessor_register, subprocessor_config_register, ISubProcessor
from functools import partial
from dlk.utils.logger import Logger
import numpy as np
import pandas as pd
import os

logger = Logger.get_logger()

@subprocessor_config_register('span_cls_relabel')
class SpanClsRelabelConfig(BaseConfig):
    default_config = {
        "_name": "span_cls_relabel",
        "config": {
            "train":{
                "input_map": {  
                     "word_ids": "word_ids",
                     "offsets": "offsets",
                     "entities_info": "entities_info",
                },
                "data_set": {                   # for different stage, this processor will process different part of data
                     "train": ['train', 'valid', 'test'],
                     "predict": ['predict'],
                },
                "output_map": {
                     "label_ids": "label_ids",
                     # for span relation extract relabel, deliver should be {"entity_id": {"start": start, "end": end}}, which the start and end should be the index of the token level
                     "processed_entities_info": "processed_entities_info",
                },
                "drop": "none", # 'longer'/'shorter'/'none', if entities is overlap, will remove by rule
                "vocab": "label_vocab", # usually provided by the "token_gather" module
                "entity_priority": [],
                "priority_trigger": 1, # if the overlap entity abs(length_a - length_b)<=priority_trigger, will trigger the entity_priority strategy, otherwise use the drop rule
                "mask_fill": -100,
                "mask_first_sent": False,  # when we use this module to resolve the MRC like SQuAD we should mask the first sentence(quesiotn) for anwering
                "null_to_zero_index": False,  # if cannot find the entity, set to point to the first(zero) index token
            },
            "extend_train": "train"
        }
    }
    """Config for SpanClsRelabel
    Config Example: 
        default_config
    """
    def __init__(self, stage, config: Dict):

        super(SpanClsRelabelConfig, self).__init__(config)
        self.config = ConfigTool.get_config_by_stage(stage, config)
        self.data_set = self.config.get('data_set', {}).get(stage, [])
        if not self.data_set:
            return
        self.word_ids = self.config['input_map']['word_ids']
        self.offsets = self.config['input_map']['offsets']
        self.mask_fill = self.config['mask_fill']
        self.entities_info = self.config['input_map']['entities_info']
        self.drop = self.config['drop']
        assert self.drop in {'none', 'longer', 'shorter'}
        self.vocab = self.config['vocab']
        self.processed_entities_info = self.config['output_map']['processed_entities_info']
        self.output_labels = self.config['output_map']['label_ids']
        self.entity_priority = {entity: priority for priority, entity in enumerate(self.config['entity_priority'])}
        self.priority_trigger = self.config['priority_trigger']
        self.mask_first_sent = self.config['mask_first_sent']
        self.null_to_zero_index = self.config['null_to_zero_index']
        self.post_check(self.config, used=[
            "drop",
            "vocab",
            "mask_first_sent",
            "input_map",
            "data_set",
            "output_map",
            "entity_priority",
            "priority_trigger",
            "null_to_zero_index",
            "mask_first_sent",
            "mask_fill"
        ])


@subprocessor_register('span_cls_relabel')
class SpanClsRelabel(ISubProcessor):
    """
    Relabel the char level entity span to token level and construct matrix
    """

    def __init__(self, stage: str, config: SpanClsRelabelConfig):
        super().__init__()
        self.stage = stage
        self.config = config
        self.vocab = None
        self.data_set = config.data_set
        if not self.data_set:
            logger.info(f"Skip 'span_cls_relabel' at stage {self.stage}")
            return

    def process(self, data: Dict)->Dict:
        """SpanClsRelabel Entry

        Args:
            data: Dict

        Returns: 
            
            relabeled data

        """

        if not self.data_set:
            return data
        # NOTE: only load once, because the vocab should not be changed in same process
        if not self.vocab:
            self.vocab = Vocabulary.load(data[self.config.vocab])
            assert self.vocab.word2idx[self.vocab.unknown] == 0, f"For span_relation_relabel, 'unknown' must be index 0, and other labels as 1...num_label"
            assert not self.vocab.pad, f"For span_relation_relabel, 'pad' must be index 0, and other labels as 1...num_label"

        for data_set_name in self.data_set:
            if data_set_name not in data['data']:
                logger.info(f'The {data_set_name} not in data. We will skip do span_cls_relabel on it.')
                continue
            data_set = data['data'][data_set_name]
            if os.environ.get('DISABLE_PANDAS_PARALLEL', 'false') != 'false':
                data_set[[self.config.output_labels,
                    self.config.processed_entities_info,
                ]] = data_set.parallel_apply(self.relabel, axis=1, result_type='expand')
            else:
                data_set[[self.config.output_labels,
                    self.config.processed_entities_info,
                ]] = data_set.apply(self.relabel, axis=1, result_type='expand')

        return data

    def find_position_in_offsets(self, position: int, offset_list: List, sub_word_ids: List, start: int, end: int, is_start: bool=False):
        """find the sub_word index which the offset_list[index][0]<=position<offset_list[index][1]

        Args:
            position: position
            offset_list: list of all tokens offsets
            sub_word_ids: word_ids from tokenizer
            start: start search index
            end: end search index
            is_start: is the position is the start of target token, if the is_start==True and cannot find return -1

        Returns: 
            the index of the offset which include position

        """
        while start<end:
            if sub_word_ids[start] is None:
                start += 1
            elif position>=offset_list[start][0] and position<offset_list[start][1]:
                return start
            elif position<offset_list[start][0]:
                if start == 1 and offset_list[0] == [0, 0]:
                    return 1
                if is_start:
                    return -1
                else:
                    return start - 1
            else:
                start += 1
        return -1

    def relabel(self, one_ins: pd.Series):
        """make token label, if use the first piece label please use the 'span_cls_firstpiece_relabel'

        Args:
            one_ins: include sentence, entity_info, offsets

        Returns: 
            labels(labels for each subtoken)
            entities_info
            processed_entities_info: for relation relabal
        """
        pre_clean_entities_info = one_ins[self.config.entities_info]
        if self.config.drop != 'none':
            pre_clean_entities_info.sort(key=lambda x: x['start'])
        offsets = one_ins[self.config.offsets]
        sub_word_ids = one_ins[self.config.word_ids]
        if not sub_word_ids:
            logger.warning(f"entity_info: {pre_clean_entities_info}, offsets: {offsets} ")

        if self.config.mask_first_sent:
            first_start = sub_word_ids.index(0)
            second_start = sub_word_ids[first_start+1:].index(0)
            mask_first_index = first_start + second_start + 2
        else:
            mask_first_index = 0 # if there is only one sentence, set to 0

        entities_info = []
        pre_end = -1
        pre_length = 0
        pre_label = ''
        for entity_info in pre_clean_entities_info:
            assert len(entity_info['labels']) == 1, f"currently we just support one label for one entity"
            if entity_info['start']<pre_end: # if overlap will remove one
                if self.config.drop == 'none':
                    pass
                elif abs(entity_info['end'] - entity_info['start'] - pre_length) <= self.config.priority_trigger:
                    pre_label_order = self.config.entity_priority.get(pre_label, 1e9)
                    label_order = self.config.entity_priority.get(entity_info['labels'][0], 1e9)
                    if label_order<pre_label_order:
                        entities_info.pop()
                    else:
                        continue
                elif self.config.drop == 'shorter':
                    if entity_info['end'] - entity_info['start'] > pre_length:
                        entities_info.pop()
                    else:
                        continue
                elif self.config.drop =='longer':
                    if entity_info['end'] - entity_info['start'] < pre_length:
                        entities_info.pop()
                    else:
                        continue
                else:
                    raise PermissionError(f"The drop method must in 'none'/'shorter'/'longer'")
            pre_label = entity_info['labels'][0]
            entities_info.append(entity_info)
            pre_end = entity_info['end']
            pre_length = entity_info['end'] - entity_info['start']
            
        offset_length = len(offsets)

        unknown_id = self.vocab.get_index(self.vocab.unknown)
        mask_matrices = np.full((offset_length, offset_length), self.config.mask_fill, dtype=np.int8)
        mask_matrices = np.tril(mask_matrices, k=-1)

        unknown_matrices = np.full((offset_length, offset_length), unknown_id, dtype=np.int8)
        unknown_matrices = np.triu(unknown_matrices, k=0)

        label_matrices = unknown_matrices + mask_matrices
        if sub_word_ids[0] is None:
            label_matrices[0, :] = self.config.mask_fill
        if sub_word_ids[-1] is None:
            label_matrices[:, -1] = self.config.mask_fill

        if mask_first_index > 0:
            label_matrices[:mask_first_index, :] = self.config.mask_fill
            label_matrices[:, :mask_first_index] = self.config.mask_fill
        processed_entities_info = []
        for entity_info in entities_info:
            if entity_info['start']==0 and entity_info['end']==0:
                start_token_index, end_token_index = 0, 0
            else:
                start_token_index = self.find_position_in_offsets(entity_info['start'], offsets, sub_word_ids, mask_first_index, offset_length, is_start=True)
                if start_token_index == -1:
                    if self.config.null_to_zero_index:
                        start_token_index, end_token_index = 0, 0
                    else:
                        logger.warning(f"cannot find the entity_info : {entity_info}, offsets: {offsets} ")
                        continue
                else:
                    end_token_index = self.find_position_in_offsets(entity_info['end']-1, offsets, sub_word_ids, start_token_index, offset_length)
                    if self.config.null_to_zero_index and end_token_index == -1:
                        start_token_index, end_token_index = 0, 0
            assert end_token_index != -1, f"entity_info: {entity_info}, offsets: {offsets}"
            label_id = self.vocab.get_index(entity_info['labels'][0])
            label_matrices[start_token_index, end_token_index] = label_id
            entity_info['sub_token_start'] = start_token_index
            entity_info['sub_token_end'] = end_token_index
            processed_entities_info.append(entity_info)

        return label_matrices, processed_entities_info
